import os
import unittest

import torch
import torch._dynamo

from torchsurv.loss.cox import neg_partial_log_likelihood as cox
from torchsurv.loss.weibull import neg_log_likelihood as weibull

# set seed for reproducibility
torch.manual_seed(42)

# Disable TorchScript JIT
os.environ["PYTORCH_JIT"] = "0"
torch._dynamo.config.suppress_errors = True


class TestTorchCompile(unittest.TestCase):
    """
    Tests using torch.compile with cox
    """

    N = 512

    def test_cox_equivalence(self):
        """
        whether the compiled version of cox evaluates to the same value
        """

        # random data and parameters
        log_hz = torch.randn(self.N)
        event = torch.randint(low=0, high=2, size=(self.N,)).bool()
        time = torch.randint(low=1, high=100, size=(self.N,))

        ccox = torch.compile(cox)  # scripted version of cox
        scox = torch.jit.script(cox)  # compiled version of cox

        loss_cox = cox(log_hz, event, time, ties_method="efron")
        loss_scox = scox(log_hz, event, time, ties_method="efron")
        loss_ccox = ccox(log_hz, event, time, ties_method="efron")

        self.assertTrue(
            torch.allclose(loss_cox, loss_scox, rtol=1e-3, atol=1e-3),
            msg="scripted failed",
        )
        self.assertTrue(
            torch.allclose(loss_cox, loss_ccox, rtol=1e-3, atol=1e-3),
            msg="compiled failed",
        )

    def test_weibull_equivalence(self):
        """
        whether the compiled version of weibull evaluates to the same value
        """

        # random data and parameters
        log_hz = torch.randn(self.N)
        event = torch.randint(low=0, high=2, size=(self.N,)).float()
        time = torch.randint(low=1, high=100, size=(self.N,))

        cweibull = torch.compile(weibull)  # scripted version of cox
        sweibull = torch.jit.script(weibull)  # compiled version of cox

        loss_weibull = weibull(log_hz, event, time)
        loss_cweibull = cweibull(log_hz, event, time)
        loss_sweibull = sweibull(log_hz, event, time)

        self.assertTrue(
            torch.allclose(loss_weibull, loss_sweibull, rtol=1e-3, atol=1e-3),
            msg="scripted failed",
        )
        self.assertTrue(
            torch.allclose(loss_weibull, loss_cweibull, rtol=1e-3, atol=1e-3),
            msg="compiled failed",
        )


if __name__ == "__main__":
    unittest.main()



import copy
import sys
import warnings
from typing import Optional, Tuple

import torch
from scipy import stats
from torchmetrics import regression

from torchsurv.tools.validate_data import validate_log_shape, validate_survival_data

__all__ = ["ConcordanceIndex"]


class ConcordanceIndex:
    """Compute the Concordance Index (C-index) for survival models."""

    def __init__(
        self,
        tied_tol: float = 1e-8,
        checks: bool = True,
    ) -> dict:
        """Initialize a ConcordanceIndex for survival class model evaluation.

        Args:
            tied_tol (float):
                Tolerance for tied risk scores.
                Defaults to 1e-8.
            checks (bool):
                Whether to perform input format checks.
                Enabling checks can help catch potential issues in the input data.
                Defaults to True.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 64
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.randn((n,))
            >>> cindex = ConcordanceIndex()
            >>> cindex(estimate, event, time)
            tensor(0.5337)
            >>> cindex.confidence_interval() # default: Noether, two_sided
            tensor([0.3251, 0.7423])
            >>> cindex.p_value(method='bootstrap', alternative='greater')
            tensor(0.2620)
        """
        self.tied_tol = tied_tol
        self.checks = checks

        # init instate attributes
        self.time = None
        self.event = None
        self.estimate = None
        self.cindex = None
        self.concordant = None
        self.discordant = None
        self.tied_risk = None
        self.weight = None
        self.weight_squared = None
        self.tmax = None

    # disable long line check for this function due to docstring. Might be better to try and do multiline math in the TeX formula
    # pylint: disable=C0301
    def __call__(
        self,
        estimate: torch.Tensor,
        event: torch.Tensor,
        time: torch.Tensor,
        weight: Optional[torch.Tensor] = None,
        tmax: Optional[torch.Tensor] = None,
        instate: bool = True,
    ) -> torch.Tensor:
        r"""Compute the Concordance Index (C-index).

        The concordance index is the probability that a model correctly predicts
        which of two comparable samples will experience an event first based on their
        estimated risk scores.

        Args:
            estimate (torch.Tensor):
                Estimated risk of event occurrence (i.e., risk score).
                Can be of shape = (n_samples,) if subject-specific risk score is time-independent,
                or of shape = (n_samples, n_samples) if subject-specific risk score is evaluated at ``time``.
            event (torch.Tensor, boolean):
                Event indicator of size n_samples (= True if event occurred).
            time (torch.Tensor, float):
                Time-to-event or censoring of size n_samples.
            weight (torch.Tensor, optional):
                Optional sample weight of size n_samples. Defaults to 1.
            tmax (torch.Tensor, optional):
                Truncation time. Defaults to None.
                ``tmax`` should be chosen such that the probability of
                being censored after time ``tmax`` is non-zero. If ``tmax`` is None, no truncation
                is performed.
            instate (bool):
                Whether to create/overwrite internal attributes states.
                Defaults to True.

        Returns:
            torch.Tensor: Concordance-index

        Note:

            The concordance index provides a global assessment of a fitted survival model over the entire observational
            period rather than focussing on the prediction for a fixed time (e.g., 10-year mortality).
            It is recommended to use AUC instead of the concordance index for such time-dependent predictions, as
            AUC is proper in this context, while the concordance index is not :cite:p:`Blanche2018`.

            For each subject :math:`i \in \{1, \cdots, N\}`, denote :math:`X_i` as the survival time and :math:`D_i` as the
            censoring time. Survival data consist of the event indicator, :math:`\delta_i=(X_i\leq D_i)`
            (argument ``event``) and the time-to-event or censoring, :math:`T_i = \min(\{ X_i,D_i \})`
            (argument ``time``).

            The risk score measures the risk (or a proxy thereof) that a subject has an event.
            The function accepts time-dependent risk score and time-independent risk score. The time-dependent risk score
            of subject :math:`i` is specified through a function :math:`q_i: [0, \infty) \rightarrow \mathbb{R}`.
            The time-independent risk score of subject :math:`i` is specified by a constant :math:`q_i`.
            The argument ``estimate`` is the estimated risk score.
            For time-dependent risk score, the argument ``estimate`` should be a tensor of shape = (N,N)
            (:math:`(i,j)` th element is :math:`\hat{q}_i(T_j)`). For time-independent risk score, the argument ``estimate``
            should be a tensor of size N (:math:`i` th element is :math:`\hat{q}_i`).

            For a pair :math:`(i,j)`, we say that the pair is comparable if the event of :math:`i` has occurred before
            the event of :math:`j`, i.e., :math:`X_i < X_j`. Given that the pair is comparable, we say that the pair is
            concordant if :math:`q_i(X_i) > q_j(X_i)`.

            The concordance index measures the probability that, for a pair of randomly selected comparable samples,
            the one that experiences an event first has a higher risk. The concordance index is defined as

            .. math::

                C = p(q_i(X_i) > q_j(X_i) \: | \: X_i < X_j)

            The default concordance index estimate is the popular nonparametric estimation proposed by :cite:t:`Harrell1996`

            .. math::

                \hat{C} = \frac{\sum_i\sum_j \delta_i \: I(T_i < T_j)\left(I \left( \hat{q}_i(T_i) > \hat{q}_j(T_i) \right) + \frac{1}{2} I\left(\hat{q}_i(T_i) = \hat{q}_j(T_i)\right)\right)}{\sum_i\sum_j \delta_i\: I\left(T_i < T_j\right)}

            When the event time are censored, the Harrell's concordance index converges to an association measure that
            involves the censoring distribution.
            To address this shortcoming, :cite:t:`Uno2011`  proposed to employ the
            inverse probability weighting technique. In this context, each subject with event time at
            :math:`t` is weighted by the inverse probability of censoring :math:`\omega(t) = 1 / \hat{D}(t)`, where
            :math:`\hat{D}(t)` is the Kaplan-Meier estimate of the censoring distribution, :math:`P(D>t)`.
            Let :math:`\omega(T_i)` be the weight associated with subject time :math:`i` (argument ``weight``).
            The concordance index estimate with weight is,

            .. math::

                \hat{C} = \frac{\sum_i\sum_j \delta_i \: \omega(T_i)^2 \: I(T_i < T_j)\left(I \left( \hat{q}_i(T_i) > \hat{q}_j(T_i) \right) + \frac{1}{2} I\left(\hat{q}_i(T_i) = \hat{q}_j(T_i)\right)\right)}{\sum_i\sum_j \delta_i \: \omega(T_i)^2\: I\left(T_i < T_j\right)}

            In the context of train/test split, the weights should be derived from the censoring distribution estimated in the training data.
            Specifically, the censoring distribution is estimated using the training set and then evaluated at the subject time within the test set.

        Examples:
            >>> from torchsurv.stats.ipcw import get_ipcw
            >>> _ = torch.manual_seed(42)
            >>> n = 64
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.randn((n,))
            >>> cindex = ConcordanceIndex()
            >>> cindex(estimate, event, time) # Harrell's c-index
            tensor(0.5337)
            >>> ipcw = get_ipcw(event, time) # ipcw at subject time
            >>> cindex(estimate, event, time, weight=ipcw) # Uno's c-index
            tensor(0.5453)

        References:

            .. bibliography::
                :filter: False

                Blanche2018
                Harrell1996
                Uno2011

        """
        # update inputs if necessary
        estimate = ConcordanceIndex._update_cindex_estimate(estimate)
        weight = ConcordanceIndex._update_weight(time, weight, tmax)

        # square the weight
        weight_squared = torch.square(weight)

        # Inputs checks
        if self.checks:
            validate_survival_data(event, time)
            validate_log_shape(estimate)

        # find comparable pairs
        comparable = self._get_comparable_and_tied_time(event, time)

        # get order index by time
        order = torch.argsort(time)

        # Initialize variables to count concordant, discordant, and tied risk pairs
        concordant, discordant, tied_risk = [], [], []

        # Initialize numerator and denominator for calculating the concordance index
        numerator, denominator = 0.0, 0.0

        # Iterate through the comparable items over time
        # (dictionary with indices that have an event occurred at time and boolean masks of comparable pair a time)
        for ind, mask in comparable.items():
            # Extract risk score, event indicator and weight for the current sample
            est_i, event_i, w_i = (
                estimate[order[ind], order[ind]],
                event[order[ind]],
                weight_squared[order[ind]],
            )

            # Extract risk score of comparable pairs and the number of comparable samples
            est_j, n = estimate[order[mask], order[ind]], mask.sum()

            # Check that the current sample is uncensored
            assert (
                event_i
            ), f"Got censored sample at index {order[ind]}, but expected uncensored"

            # Identify tied pairs based on a tolerance (ties)
            ties = (
                torch.absolute(est_j.float() - est_i.float()) <= self.tied_tol
            )  # ties
            n_ties = ties.sum()

            # Identify concordant pairs
            con = est_j < est_i
            n_con = con[~ties].sum()

            # Update numerator and denominator for concordance index calculation
            numerator += w_i * n_con + 0.5 * w_i * n_ties
            denominator += w_i * n

            # Update counts for tied, concordant, and discordant pairs
            tied_risk.append(w_i * n_ties)
            concordant.append(w_i * n_con)
            discordant.append(w_i * n - w_i * n_con - w_i * n_ties)

        # Create/overwrite internal attributes states
        if instate:
            self.time = time
            self.event = event
            self.estimate = estimate
            self.cindex = numerator / denominator
            self.concordant = torch.stack(concordant, dim=0)
            self.discordant = torch.stack(discordant, dim=0)
            self.tied_risk = torch.stack(tied_risk, dim=0)
            self.weight = weight
            self.weight_squared = weight_squared
            self.tmax = tmax

        return numerator / denominator  # cindex

    def confidence_interval(
        self,
        method: str = "noether",
        alpha: float = 0.05,
        alternative: str = "two_sided",
        n_bootstraps: int = 999,
    ) -> torch.Tensor:
        """Compute the confidence interval of the Concordance index.

        This function calculates either the pointwise confidence interval or the bootstrap
        confidence interval for the concordance index. The pointwise confidence interval is computed
        assuming that the concordance index is normally distributed and using the standard error estimated with either the Noether or
        the conservative method :cite:p:`Pencina2004`.
        The bootstrap confidence interval is constructed based on the distribution of bootstrap samples.

        Args:
            method (str):
                Method for computing confidence interval. Defaults to "noether".
                Must be one of the following: "noether", "conservative", "bootstrap".
            alpha (float):
                Significance level. Defaults to 0.05.
            alternative (str):
                Alternative hypothesis. Defaults to "two_sided".
                Must be one of the following: "two_sided", "greater", "less".
            n_bootstraps (int):
                Number of bootstrap samples. Defaults to 999.
                Ignored if ``method`` is not "bootstrap".

        Returns:
            torch.Tensor([lower,upper]):
                Lower and upper bounds of the confidence interval.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 64
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.randn((n,))
            >>> cindex = ConcordanceIndex()
            >>> cindex(estimate, event, time)
            tensor(0.5337)
            >>> cindex.confidence_interval() # default: Noether, two_sided
            tensor([0.3251, 0.7423])
            >>> cindex.confidence_interval(method="bootstrap", alternative="greater")
            tensor([0.4459, 1.0000])
            >>> cindex.confidence_interval(method="conservative", alternative="less")
            tensor([0.0000, 0.7558])

        References:
            .. bibliography::
                :filter: False

                Pencina2004
        """

        assert isinstance(method, str)
        assert isinstance(alternative, str)
        assert (
            hasattr(self, "cindex") and self.cindex is not None
        ), "Error: Please calculate the concordance index using `ConcordanceIndex()` before calling `confidence_interval()`."

        if alternative not in ["less", "greater", "two_sided"]:
            raise ValueError(
                f"'alternative' {alternative} must be one of ['less', 'greater', 'two_sided']."
            )

        if method == "noether":
            conf_int = self._confidence_interval_noether(alpha, alternative)
        elif method == "bootstrap":
            conf_int = self._confidence_interval_bootstrap(
                alpha, alternative, n_bootstraps
            )
        elif method == "conservative":
            conf_int = self._confidence_interval_conservative(alpha, alternative)
        else:
            raise ValueError(
                f"Method {method} not implemented. Please choose either 'noether', 'conservative' or 'bootstrap'."
            )
        return conf_int

    def p_value(
        self,
        method: str = "noether",
        alternative: str = "two_sided",
        n_bootstraps: int = 999,
    ) -> torch.Tensor:
        """Perform one-sample hypothesis test on the Concordance index.

        This function calculates either the pointwise p-value or the bootstrap p-value
        for testing the null hypothesis that the concordance index is equal to 0.5.
        The pointwise p-value is computed assuming that the concordance index is normally distributed and using the
        standard error estimated using the Noether's method :cite:p:`Pencina2004`.
        The bootstrap p-value is derived by permuting risk predictions to estimate
        the sampling distribution under the null hypothesis.

        Args:
            method (str):
                Method for computing p-value. Defaults to "noether".
                Must be one of the following: "noether", "bootstrap".
            alternative (str):
                Alternative hypothesis. Defaults to "two_sided".
                Must be one of the following: "two_sided" (concordance index is not equal to 0.5),
                "greater" (concordance index is greater than 0.5), "less" (concordance index is less than 0.5).
            n_bootstraps (int):
                Number of bootstrap samples. Defaults to 999.
                Ignored if ``method`` is not "bootstrap".

        Returns:
            torch.Tensor: p-value of statistical test.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 64
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.randn((n,))
            >>> cindex = ConcordanceIndex()
            >>> cindex(estimate, event, time)
            tensor(0.5337)
            >>> cindex.p_value() # default: Noether, two_sided
            tensor(0.7516)
            >>> cindex.p_value(method="bootstrap", alternative="greater")
            tensor(0.2620)

        """

        assert (
            hasattr(self, "cindex") and self.cindex is not None
        ), "Error: Please calculate the concordance index using `ConcordanceIndex()` before calling `p_value()`."

        if alternative not in ["less", "greater", "two_sided"]:
            raise ValueError(
                "'alternative' parameter must be one of ['less', 'greater', 'two_sided']."
            )

        if method == "noether":
            pvalue = self._p_value_noether(alternative)
        elif method == "bootstrap":
            pvalue = self._p_value_bootstrap(alternative, n_bootstraps)
        else:
            raise ValueError(
                f"Method {method} not implemented. Please choose either 'noether' or 'bootstrap'."
            )
        return pvalue

    def compare(
        self, other, method: str = "noether", n_bootstraps: int = 999
    ) -> torch.Tensor:
        """Compare two Concordance indices.

        This function compares two concordance indices computed on the
        same data with different risk scores. The statistical hypotheses are
        formulated as follows, null hypothesis: cindex1 = cindex2 and alternative
        hypothesis: cindex1 > cindex2.
        The statistical test is either a Student t-test for dependent samples or
        a two-sample bootstrap test. The Student t-test assumes that the concordance index is normally distributed
        and uses standard errors estimated with the Noether's method :cite:p:`Pencina2004`.

        Args:
            other (ConcordanceIndex):
                Another instance of the ConcordanceIndex class representing cindex2.
            method (str):
                Statistical test used to perform the hypothesis test. Defaults to "noether".
                Must be one of the following: "noether", "bootstrap".
            n_bootstraps (int):
                Number of bootstrap samples. Defaults to 999.
                Ignored if ``method`` is not "bootstrap".

        Returns:
            torch.Tensor: p-value of the statistical test.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 64
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> cindex1 = ConcordanceIndex()
            >>> cindex1(torch.randn((n,)), event, time)
            tensor(0.5337)
            >>> cindex2 = ConcordanceIndex()
            >>> cindex2(torch.randn((n,)), event, time)
            tensor(0.5047)
            >>> cindex1.compare(cindex2) # default: Noether
            tensor(0.4267)
            >>> cindex1.compare(cindex2, method = "bootstrap")
            tensor(0.3620)

        """

        assert isinstance(other, ConcordanceIndex)
        assert isinstance(method, str)

        assert (
            hasattr(self, "cindex") and self.cindex is not None
        ), "Error: Please calculate the concordance index using `ConcordanceIndex()` before calling `compare()`."

        # assert that the same data were used to compute the two c-index
        if torch.any(self.event != other.event) or torch.any(self.time != other.time):
            raise ValueError(
                "Mismatched survival data: 'time' and 'event' should be the same for both concordance index computations."
            )

        if self.tmax != other.tmax:
            raise ValueError(
                "Mismatched truncation time: 'tmax' should be the same for both concordance index computations."
            )

        if method == "noether":
            pvalue = self._compare_noether(other)
        elif method == "bootstrap":
            pvalue = self._compare_bootstrap(other, n_bootstraps)
        else:
            raise ValueError(
                f"Method {method} not implemented. Please choose either 'noether' or 'bootstrap'."
            )
        return pvalue

    def _confidence_interval_noether(
        self, alpha: float, alternative: str
    ) -> torch.Tensor:
        """Confidence interval of Concordance index assuming that the concordance index
        is normally distributed and using standard errors estimated using Noether's method.
        """

        alpha = alpha / 2 if alternative == "two_sided" else alpha

        cindex_se = self._concordance_index_se()

        if cindex_se > 0:
            ci = (
                -torch.distributions.normal.Normal(0, 1).icdf(
                    torch.tensor(alpha, device=self.cindex.device)
                )
                * cindex_se
            )
            lower = torch.max(
                torch.tensor(0.0, device=self.cindex.device), self.cindex - ci
            )
            upper = torch.min(
                torch.tensor(1.0, device=self.cindex.device), self.cindex + ci
            )

            if alternative == "less":
                lower = torch.tensor(0.0, device=self.cindex.device)
            elif alternative == "greater":
                upper = torch.tensor(1.0, device=self.cindex.device)
        else:
            raise ValueError(
                "The standard error of the concordance index must be a positive value."
            )

        return torch.stack([lower, upper], dim=0)

    # pylint: disable=invalid-name
    def _confidence_interval_conservative(
        self, alpha: float, alternative: str
    ) -> torch.Tensor:
        """Confidence interval of Concordance index assuming that the concordance index
        is normally distributed and using the conservative method.
        """
        alpha = alpha / 2 if alternative == "two_sided" else alpha

        N = torch.sum(self.weight)

        pc = (1 / (N * (N - 1))) * torch.sum(self.concordant)
        pd = (1 / (N * (N - 1))) * torch.sum(self.discordant)

        w = (
            (
                torch.distributions.normal.Normal(0, 1).icdf(
                    torch.tensor(alpha, device=self.cindex.device)
                )
                ** 2
            )
            * 2
        ) / (N * (pc + pd))

        ci = torch.sqrt(w**2 + 4 * w * self.cindex * (1 - self.cindex)) / (
            2 * (1 + w)
        )
        point = (w + 2 * self.cindex) / (2 * (1 + w))

        lower = point - ci
        upper = point + ci

        if alternative == "less":
            lower = torch.tensor(0.0, device=self.cindex.device)
        elif alternative == "greater":
            upper = torch.tensor(1.0, device=self.cindex.device)
        return torch.stack([lower, upper])

    def _confidence_interval_bootstrap(
        self, alpha: float, alternative: str, n_bootstraps: int
    ) -> torch.Tensor:
        """Bootstrap confidence interval of the Concordance index using
        Efron's percentile method.

        References:
            Efron, Bradley; Tibshirani, Robert J. (1993).
                An introduction to the bootstrap, New York: Chapman & Hall, software.
        """

        # c-index bootstraps given bootstrap distribution
        cindex_bootstrap = self._bootstrap_cindex(
            metric="confidence_interval", n_bootstraps=n_bootstraps
        )

        # obtain confidence interval
        if alternative == "two_sided":
            lower, upper = torch.quantile(
                cindex_bootstrap,
                torch.tensor([alpha / 2, 1 - alpha / 2], device=self.cindex.device),
            )
        elif alternative == "less":
            upper = torch.quantile(
                cindex_bootstrap, torch.tensor(1 - alpha, device=self.cindex.device)
            )
            lower = torch.tensor(0.0, device=self.cindex.device)
        elif alternative == "greater":
            lower = torch.quantile(
                cindex_bootstrap, torch.tensor(alpha, device=self.cindex.device)
            )
            upper = torch.tensor(1.0, device=self.cindex.device)

        return torch.stack([lower, upper])

    def _p_value_noether(self, alternative, null_value: float = 0.5) -> torch.Tensor:
        """p-value for a one-sample hypothesis test of the Concordance index
        assuming that the concordance index is normally distributed and using standard
        errors estimated with Noether's method.
        """

        cindex_se = self._concordance_index_se()

        # get p-value
        if cindex_se > 0:
            p = torch.distributions.normal.Normal(0, 1).cdf(
                (self.cindex - null_value) / cindex_se
            )
            if alternative == "two_sided":
                if self.cindex >= torch.tensor(0.5):
                    p = torch.tensor(1.0) - p
                p *= torch.tensor(2.0)
            elif alternative == "greater":
                p = torch.tensor(1.0) - p
        else:
            raise ValueError(
                "The standard error of concordance index must be a positive value."
            )

        return p

    def _p_value_bootstrap(self, alternative, n_bootstraps) -> torch.Tensor:
        """p-value for a one-sample hypothesis test of the Concordance Index using
        permutation of risk prediction to estimate sampling distribution under the null
        hypothesis.
        """

        # c-index bootstraps given null distribution cindex = 0.5
        cindex0 = self._bootstrap_cindex(metric="p_value", n_bootstraps=n_bootstraps)

        # Derive p-value
        p = (torch.tensor(1) + torch.sum(cindex0 <= self.cindex)) / torch.tensor(
            n_bootstraps + 1
        )
        if alternative == "two_sided":
            if self.cindex >= torch.tensor(0.5):
                p = torch.tensor(1.0) - p
            p *= torch.tensor(2.0)
            p = torch.min(
                torch.tensor(1.0, device=self.cindex.device), p
            )  # in case very small bootstrap sample size is used
        elif alternative == "greater":
            p = torch.tensor(1.0) - p

        return p

    # pylint: disable=W0212
    def _compare_noether(self, other):
        """Student t-test for dependent samples given Noether's standard error to
        compare two concordance indices.
        """

        # sample size
        N = sum(self.weight)

        # compute noether standard error
        cindex1_se = self._concordance_index_se()
        cindex2_se = other._concordance_index_se()

        # Suppress the specific warning
        warnings.filterwarnings(
            "ignore",
            message="Metric `SpearmanCorrcoef` will save all targets and predictions in the buffer. For large datasets, this may lead to large memory footprint.",
            category=UserWarning,
        )

        # compute spearman correlation between risk prediction
        corr = regression.SpearmanCorrCoef()(
            self.estimate.reshape(-1), other.estimate.reshape(-1)
        )

        # check for perfect positive monotonic relationship between two variables
        if 1 - torch.abs(corr) < 1e-15:
            return 1.0

        # compute t-stat
        t_stat = (self.cindex - other.cindex) / torch.sqrt(
            cindex1_se**2 + cindex2_se**2 - 2 * corr * cindex1_se * cindex2_se
        )

        # return p-value
        return torch.tensor(
            1 - stats.t.cdf(t_stat, df=N - 1),
            dtype=self.cindex.dtype,
            device=self.cindex.device,
        )  # student-t cdf not available on torch

    def _compare_bootstrap(self, other, n_bootstraps):
        """Bootstrap two-sample test to compare two concordance indices."""

        # c-index bootstraps given null hypothesis that cindex1
        # and cindex2 come from the same distribution
        cindex1_null = self._bootstrap_cindex(
            metric="compare", other=other, n_bootstraps=n_bootstraps
        )
        cindex2_null = self._bootstrap_cindex(
            metric="compare", other=other, n_bootstraps=n_bootstraps
        )

        # bootstrapped test statistics
        t_boot = cindex1_null - cindex2_null

        # observed test statistics
        t_obs = self.cindex - other.cindex

        # return p-value
        return torch.tensor(1) - (
            torch.tensor(1) + torch.sum(t_boot <= t_obs)
        ) / torch.tensor(n_bootstraps + 1)

    def _concordance_index_se(self):
        """Standard error of concordance index using Noether's method."""

        N = sum(self.weight)

        pc = (1 / (N * (N - 1))) * torch.sum(self.concordant)
        pd = (1 / (N * (N - 1))) * torch.sum(self.discordant)
        pcc = (1 / (N * (N - 1) * (N - 2))) * torch.sum(
            self.concordant * (self.concordant - 1)
        )
        pdd = (1 / (N * (N - 1) * (N - 2))) * torch.sum(
            self.discordant * (self.discordant - 1)
        )
        pcd = (1 / (N * (N - 1) * (N - 2))) * torch.sum(
            self.concordant * self.discordant
        )
        varp = (4 / (pc + pd) ** 4) * (
            pd**2 * pcc - 2 * pc * pd * pcd + pc**2 * pdd
        )

        return torch.sqrt(varp / N)

    def _bootstrap_cindex(
        self, metric: str, n_bootstraps: int, other=None
    ) -> torch.Tensor:
        """Compute bootstrap samples of the Concordance Index (C-index).

        Args:
            metric (str): Must be one of the following: "p_value", "confidence_interval", "compare".
                If "p_value", computes bootstrap samples of the concordance index given the sampling distribution
                under the null hypothesis (c-index = 0.5). If "confidence_interval", computes bootstrap
                samples of the c-index given the data distribution. If "compare", computes
                bootstrap samples of the c-index given the sampling distribution under the comparison test
                null hypothesis (c-index1 = cindex2).
            n_bootstraps (int): Number of bootstrap samples.
            other (optional, ConcordanceIndex):
                 Another instance of the ConcordanceIndex class representing cindex2.
                Only required for the ``metric`` is "compare".


        Returns:
            torch.Tensor: bootstrap samples of Concordance index.
        """

        # Initiate empty list to store concordance index
        cindexes = []

        # Get the bootstrap samples of concordance index
        for _ in range(n_bootstraps):
            if (
                metric == "p_value"
            ):  # bootstrap samples given null distribution (cindex = 0.5)
                estimate = copy.deepcopy(self.estimate)
                estimate = estimate[
                    torch.randperm(len(estimate)), :
                ]  # Shuffle estimate
                cindexes.append(
                    self(
                        estimate,
                        self.event,
                        self.time,
                        self.weight,
                        self.tmax,
                        instate=False,
                    )
                )  # Run Concordance index, without saving internal state
            elif (
                metric == "confidence_interval"
            ):  # bootstrap samples given data distribution
                index = torch.randint(
                    low=0,
                    high=len(self.event),
                    size=(
                        len(
                            self.event,
                        ),
                    ),
                )
                cindexes.append(
                    self(  # sample with replacement from sample
                        self.estimate[index, :][:, index],
                        self.event[index],
                        self.time[index],
                        self.weight,
                        self.tmax,
                        instate=False,
                    )
                )  # Run Concordance index, without saving internal state
            elif (
                metric == "compare"
            ):  # bootstrap samples given null distribution (cindex1 = cindex2)
                index = torch.randint(
                    low=0,
                    high=len(self.event) * 2,
                    size=(
                        len(
                            self.event,
                        ),
                    ),
                )
                estimate = torch.cat(
                    (
                        torch.cat((self.estimate, self.estimate), dim=1),
                        torch.cat((other.estimate, other.estimate), dim=1),
                    ),
                    dim=0,
                )  # create n_samples*2, n_samples*2 tensor
                cindexes.append(
                    self(  # sample with replacement from pooled sample
                        estimate[index, :][:, index],
                        torch.cat((self.event, other.event))[index],
                        torch.cat((self.time, other.time))[index],
                        torch.cat((self.weight, other.weight))[index],
                        self.tmax,
                        instate=False,
                    )
                )

        return torch.stack(cindexes)

    @staticmethod
    def _get_comparable_and_tied_time(
        event,
        time,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Identify comparable pairs and count tied time pairs.
        The function iterates through the sorted time points to identify comparable samples
        (those with the same time point) and count the number of tied time points.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]:
                - comparable (torch.Tensor): Dictionary containing indices as keys
                and boolean masks as values, indicating which samples are comparable.
                - tied_time (torch.Tensor): Number of tied time points.

        Note:
        - 'comparable' dictionary maps indices to boolean masks, where True indicates
        that the corresponding sample is comparable to others at the same time point.
        - 'tied_time' counts the total number of tied time points.

        """
        # Number of samples
        n_samples = len(time)

        # Sort indices based on time
        order = torch.argsort(time)

        # Initialize dictionary to store comparable samples
        comparable = {}

        # Initialize count
        tied_time = 0

        # Initialize index for storing unique values
        i = 0

        # Iterate through the sorted time points
        while i < n_samples - 1:
            time_i = time[order[i]]

            # Find the range of samples with the same time point
            start = i + 1
            end = start
            while end < n_samples and time[order[end]] == time_i:
                end += 1

            # check for tied event times
            event_at_same_time = event[order[i:end]]
            censored_at_same_time = torch.logical_not(event_at_same_time)

            # Iterate through the sample with the same time point
            for j in range(i, end):
                # If event occurred at time
                if event[order[j]]:
                    # Create a boolean mask for comparability
                    mask = torch.zeros_like(time).bool()
                    mask[end:] = True

                    # Store the comparability mask for the event
                    mask[i:end] = censored_at_same_time
                    comparable[j] = mask
                    tied_time += censored_at_same_time.sum()
            i = end

        if len(comparable) == 0:
            raise ValueError("No comparable pairs, denominator is 0.")

        return comparable

    @staticmethod
    def _update_weight(
        time: torch.Tensor,
        weight: torch.Tensor,
        tmax: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Obtain subject-specific weight."""

        # If weight is not provided, use ones instead.
        weight_updated = torch.zeros_like(time)

        # If tmax is provided, truncate time after tmax (mask = False)
        masks = torch.ones_like(time).bool() if tmax is None else time < tmax
        weight_updated[masks] = (
            torch.tensor(1.0, dtype=weight_updated.dtype, device=time.device)
            if weight is None
            else weight[masks]
        )

        return weight_updated

    @staticmethod
    def _update_cindex_estimate(estimate: torch.Tensor) -> torch.Tensor:
        # Ensure estimate is (n_samples, n_samples) shape
        if estimate.ndim == 1:
            estimate = estimate.unsqueeze(1)

        if estimate.shape[1] == 1:
            estimate = estimate.expand((estimate.shape[0], estimate.shape[0]))

        return estimate


if __name__ == "__main__":
    import doctest

    # Run doctest
    results = doctest.testmod()
    if results.failed == 0:
        print("All tests passed.")
    else:
        print("Some doctests failed.")
        sys.exit(1)



import itertools
import sys
from typing import Tuple

import torch

from torchsurv.tools.validate_data import validate_survival_data

__all__ = [
    "KaplanMeierEstimator",
]


class KaplanMeierEstimator:
    """Kaplan-Meier estimate of survival or censoring distribution for right-censored data :cite:p:`Kaplan1958`."""

    def __call__(
        self,
        event: torch.Tensor,
        time: torch.Tensor,
        censoring_dist: bool = False,
        check: bool = True,
    ):
        """Initialize Kaplan Meier estimator.

        Args:
            event (torch.tensor, bool):
                Event indicator of size n_samples (= True if event occurred).
            time (torch.tensor, float):
                Time-to-event or censoring of size n_samples.
            censoring_dist (bool, optional):
                If False, returns the Kaplan-Meier estimate of the survival distribution.
                If True, returns the Kaplan-Meier estimate of the censoring distribution.
                Defaults to False.
            check (bool):
                Whether to perform input format checks.
                Enabling checks can help catch potential issues in the input data.
                Defaults to True.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 32
            >>> time = torch.randint(low=0, high=8, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> s = KaplanMeierEstimator() # estimate survival distribution
            >>> s(event, time)
            >>> s.km_est
            tensor([1.0000, 1.0000, 0.8214, 0.7143, 0.6391, 0.6391, 0.5113, 0.2556])
            >>> c = KaplanMeierEstimator() # estimate censoring distribution
            >>> c(event, time, censoring_dist = True)
            >>> c.km_est
            tensor([0.9688, 0.8750, 0.8750, 0.8312, 0.6357, 0.4890, 0.3667, 0.0000])

        References:

        .. bibliography::
            :filter: False

            Kaplan1958
        """

        # create attribute state
        # pylint: disable=attribute-defined-outside-init
        self.event = event
        self.time = time

        # Check input validity if required
        if check:
            validate_survival_data(event, time)

        # Compute the counts of events, censorings, and the number at risk at each unique time
        uniq_times, n_events, n_at_risk, n_censored = self._compute_counts()

        # If 'censoring_dist' is True, estimate the censoring distribution instead of the survival distribution
        if censoring_dist:
            n_at_risk -= n_events
            n_events = n_censored

        # Compute the Kaplan-Meier estimator
        ratio = torch.where(
            n_events != 0,  # Check if the number of events is not equal to zero
            n_events
            / n_at_risk,  # Element-wise division when the number of events is not zero
            torch.zeros_like(
                n_events, dtype=torch.float
            ),  # Set to zero when the number of events is zero to avoid division by zero
        )
        values = (
            1.0 - ratio
        )  # Compute the survival (or censoring) probabilities at each unique time
        y = torch.cumprod(
            values, dim=0
        )  # Cumulative product to get the Kaplan-Meier estimator

        # Keep track of the unique times and Kaplan-Meier estimator values
        self.time = uniq_times
        self.km_est = y

    def plot_km(self, ax=None, **kwargs):
        """Plot the Kaplan-Meier estimate of the survival distribution.

        Args:
            ax (matplotlib.axes.Axes, optional):
                The axes to plot the Kaplan-Meier estimate.
                If None, a new figure and axes are created.
                Defaults to None.
            **kwargs:
                Additional keyword arguments to pass to the plot function.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 32
            >>> time = torch.randint(low=0, high=8, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> km = KaplanMeierEstimator()
            >>> km(event, time)
            >>> km.plot_km()


        """
        import matplotlib.pyplot as plt  # pylint: disable=import-outside-toplevel

        if ax is None:
            _, ax = plt.subplots()

        ax.step(self.time, self.km_est, where="post", **kwargs)
        ax.set_xlabel("Time")
        ax.set_ylabel("Survival Probability")
        ax.set_title("Kaplan-Meier Estimate")

    def predict(self, new_time: torch.Tensor) -> torch.Tensor:
        """Predicts the Kaplan-Meier estimate on new time points.
        If the new time points do not match any times used to fit, the left-limit is used.

        Args:
            new_time (torch.tensor):
                New time points at which to predict the Kaplan-Meier estimate.

        Returns:
            Kaplan-Meier estimate evaluated at ``new_time``.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 8
            >>> time = torch.randint(low=1, high=10, size=(n * 4,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n * 4,)).bool()
            >>> km = KaplanMeierEstimator()
            >>> km(event, time)
            >>> km.predict(torch.randint(low=0, high=10, size=(n,))) # predict survival distribution
            tensor([1.0000, 0.9062, 0.8700, 1.0000, 0.9062, 0.9062, 0.4386, 0.0000])

        """

        # add probability of 1 of survival before time 0
        ref_time = torch.cat((-torch.tensor([torch.inf]), self.time), dim=0)
        km_est_ = torch.cat((torch.ones(1), self.km_est))

        # Check if newtime is beyond the last observed time point
        extends = new_time > torch.max(ref_time)
        if km_est_[torch.argmax(ref_time)] > 0 and extends.any():
            # pylint: disable=consider-using-f-string
            raise ValueError(
                "Cannot predict survival/censoring distribution after the largest observed training event time point: {}".format(
                    ref_time[-1].item()
                )
            )

        # beyond last time point is zero probability
        km_pred = torch.zeros_like(new_time, dtype=km_est_.dtype)
        km_pred[extends] = 0.0

        # find new time points that match train time points
        idx = torch.searchsorted(ref_time, new_time[~extends], side="left")

        # For non-exact matches, take the left limit (shift the index to the left)
        eps = torch.finfo(ref_time.dtype).eps
        idx[torch.abs(ref_time[idx] - new_time[~extends]) >= eps] -= 1

        # predict
        km_pred[~extends] = km_est_[idx]

        return km_pred

    def print_survival_table(self):
        """Prints the survival table with the unique times and Kaplan-Meier estimates.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 32
            >>> time = torch.randint(low=0, high=8, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> s = KaplanMeierEstimator()
        """
        # Print header
        print("Time\tSurvival")
        print("-" * 16)

        # Print unique times and Kaplan-Meier estimates
        for t, y in zip(self.time, self.km_est):
            print(f"{t:.2f}\t{y:.4f}")

    def _compute_counts(
        self,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Compute the counts of events, censorings and risk set at ``time``.

        Returns: Tuple[torch.tensor, torch.tensor, torch.tensor, torch.tensor]
            unique times
            number of events at unique times
            number at risk at unique times
            number of censored at unique times
        """
        # Get the number of samples
        n_samples = len(self.event)

        # Sort indices based on time
        order = torch.argsort(self.time, dim=0)

        # Initialize arrays to store unique times, event counts, and total counts
        uniq_times = torch.empty_like(self.time)
        uniq_events = torch.empty_like(self.time, dtype=torch.long)
        uniq_counts = torch.empty_like(self.time, dtype=torch.long)

        # Group indices by unique time values
        groups = itertools.groupby(
            range(len(self.time)), key=lambda i: self.time[order[i]]
        )

        # Initialize index for storing unique values
        j = 0

        # Iterate through unique times
        for _, group_indices in groups:
            group_indices = list(group_indices)

            # Count events and total occurrences
            count_event = sum(self.event[order[i]].item() for i in group_indices)
            count = len(group_indices)

            # Store unique time, event count, and total count
            uniq_times[j] = self.time[order[group_indices[0]]].item()
            uniq_events[j] = count_event
            uniq_counts[j] = count
            j += 1

        # Extract valid values based on the index
        times = uniq_times[:j]
        n_events = uniq_events[:j]
        total_count = uniq_counts[:j]
        n_censored = total_count - n_events

        # Offset cumulative sum by one to get the number at risk
        n_at_risk = n_samples - torch.cumsum(
            torch.cat([torch.tensor([0]), total_count]), dim=0
        )

        return times, n_events, n_at_risk[:-1], n_censored


if __name__ == "__main__":
    import doctest

    # Run doctest
    results = doctest.testmod()
    if results.failed == 0:
        print("All tests passed.")
    else:
        print("Some doctests failed.")
        sys.exit(1)



# global modules
import json
import unittest

import numpy as np
import torch
from sksurv.nonparametric import CensoringDistributionEstimator
from utils import DataBatchContainer

# Local modules
from torchsurv.stats.ipcw import get_ipcw

# Load the benchmark cox log likelihoods from R
with open("tests/benchmark_data/benchmark_ipcw.json") as file:
    benchmark_ipcws = json.load(file)

# set seed for reproducibility
torch.manual_seed(42)
np.random.seed(42)


class TestIPCW(unittest.TestCase):
    """
    List of packages compared
        - pec (R)
        - sksurv (Python)
    """

    def test_ipcw_real_data(self):
        """test ipcw values on lung and gbsg data"""

        for benchmark_ipcw in benchmark_ipcws:
            event = torch.tensor(benchmark_ipcw["status"]).bool()
            time = torch.tensor(benchmark_ipcw["time"], dtype=torch.float32)
            new_time = torch.tensor(benchmark_ipcw["times"], dtype=torch.float32)

            # ipcw evaluated at subject time
            ipcw_subject_time = get_ipcw(event, time)
            ipcw_subject_time_pec = np.array(benchmark_ipcw["ipcw_subjectimes"])

            # ipcw evaluated at new time
            ipcw_new_time = get_ipcw(event, time, new_time)
            ipcw_new_time_pec = np.array(benchmark_ipcw["ipcw_times"])

            self.assertTrue(
                np.allclose(
                    ipcw_subject_time.numpy(),
                    ipcw_subject_time_pec,
                    rtol=1e-4,
                    atol=1e-8,
                )
            )

            self.assertTrue(
                np.allclose(
                    ipcw_new_time.numpy(), ipcw_new_time_pec, rtol=1e-4, atol=1e-8
                )
            )

    def test_ipcw_simulated_data(self):
        """test ipcw on simulated batches including edge cases"""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=20,
            flags_to_set=[
                "train_ties_time_event",
                "test_ties_time_event",
                "train_ties_time_censoring",
                # "test_ties_time_censoring",
                "train_ties_time_event_censoring",
                "test_ties_time_event_censoring",
                "test_no_censoring",
                #'train_no_censoring', concordance_index_ipcw from sksurv fails
                "test_event_at_last_time",
                "ties_score_events",
                "ties_score_event_censoring",
                "ties_score_censoring",
            ],
        )
        for _, batch in enumerate(batch_container.batches):
            (
                train_time,
                train_event,
                test_time,
                test_event,
                _,
                _,
                y_train_array,
                y_test_array,
                _,
                _,
            ) = batch

            # ipcw
            ipcw = get_ipcw(train_event, train_time, test_time)

            # sksurv imposes survival data (event and time) for ipcw prediction
            # instead of just time. And then force icpw to be 0 if event == False
            ipcw[test_event == False] = 0.0

            # ipcw with sksurv
            cens = CensoringDistributionEstimator()
            cens.fit(y_train_array)
            ipcw_sksurv = cens.predict_ipcw(y_test_array)

            self.assertTrue(
                np.all(np.isclose(ipcw.numpy(), ipcw_sksurv, rtol=1e-4, atol=1e-8))
            )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=C0103
# pylint: disable=C0301

import sys
import warnings

import torch

from torchsurv.tools.validate_data import validate_loss

__all__ = [
    "_partial_likelihood_cox",
    "_partial_likelihood_efron",
    "_partial_likelihood_breslow",
    "neg_partial_log_likelihood",
]


def _partial_likelihood_cox(
    log_hz_sorted: torch.Tensor,
    event_sorted: torch.Tensor,
) -> torch.Tensor:
    """Calculate the partial log likelihood for the Cox proportional hazards model
    in the absence of ties in event time.
    """
    log_hz_flipped = log_hz_sorted.flip(0)
    log_denominator = torch.logcumsumexp(log_hz_flipped, dim=0).flip(0)
    return (log_hz_sorted - log_denominator)[event_sorted]


def _partial_likelihood_efron(
    log_hz_sorted: torch.Tensor,
    event_sorted: torch.Tensor,
    time_sorted: torch.Tensor,
    time_unique: torch.Tensor,
) -> torch.Tensor:
    """Calculate the partial log likelihood for the Cox proportional hazards model
    using Efron's method to handle ties in event time.
    """
    J = len(time_unique)

    H = [
        torch.where((time_sorted == time_unique[j]) & (event_sorted == 1))[0]
        for j in range(J)
    ]
    R = [torch.where(time_sorted >= time_unique[j])[0] for j in range(J)]

    # Calculate the length of each element in H and store it in a tensor
    m = torch.tensor([len(h) for h in H])

    # Create a boolean tensor indicating whether each element in H has a length greater than 0
    include = torch.tensor([len(h) > 0 for h in H])

    log_nominator = torch.stack([torch.sum(log_hz_sorted[h]) for h in H])

    denominator_naive = torch.stack([torch.sum(torch.exp(log_hz_sorted[r])) for r in R])
    denominator_ties = torch.stack([torch.sum(torch.exp(log_hz_sorted[h])) for h in H])

    log_denominator_efron = torch.zeros(J, device=log_hz_sorted.device)
    for j in range(J):
        mj = int(m[j].item())
        for l in range(1, mj + 1):
            log_denominator_efron[j] += torch.log(
                denominator_naive[j] - (l - 1) / float(m[j]) * denominator_ties[j]
            )
    return (log_nominator - log_denominator_efron)[include]


def _partial_likelihood_breslow(
    log_hz_sorted: torch.Tensor,
    event_sorted: torch.Tensor,
    time_sorted: torch.Tensor,
):
    """
    Compute the partial likelihood using Breslow's method for Cox proportional hazards model.

    Args:
        log_hz_sorted (torch.Tensor): Log hazard rates sorted by time.
        event_sorted (torch.Tensor): Binary tensor indicating if the event occurred (1) or was censored (0), sorted by time.
        time_sorted (torch.Tensor): Event or censoring times sorted in ascending order.

    Returns:
        torch.Tensor: The partial likelihood for the observed events.
    """
    N = len(time_sorted)
    R = [torch.where(time_sorted >= time_sorted[i])[0] for i in range(N)]
    log_denominator = torch.stack(
        [torch.logsumexp(log_hz_sorted[R[i]], dim=0) for i in range(N)]
    )

    return (log_hz_sorted - log_denominator)[event_sorted]


def neg_partial_log_likelihood(
    log_hz: torch.Tensor,
    event: torch.Tensor,
    time: torch.Tensor,
    ties_method: str = "efron",
    reduction: str = "mean",
    checks: bool = True,
) -> torch.Tensor:
    r"""Compute the negative of the partial log likelihood for the Cox proportional hazards model.

    Args:
        log_hz (torch.Tensor, float):
            Log relative hazard of length n_samples.
        event (torch.Tensor, bool):
            Event indicator of length n_samples (= True if event occurred).
        time (torch.Tensor):
            Time-to-event or censoring of length n_samples.
        ties_method (str):
            Method to handle ties in event time. Defaults to "efron".
            Must be one of the following: "efron", "breslow".
        reduction (str):
            Method to reduce losses. Defaults to "mean".
            Must be one of the following: "sum", "mean".
        checks (bool):
            Whether to perform input format checks.
            Enabling checks can help catch potential issues in the input data.
            Defaults to True.

    Returns:
        (torch.tensor, float):
            Negative of the partial log likelihood.

    Note:
        For each subject :math:`i \in \{1, \cdots, N\}`, denote :math:`X_i` as the survival time and :math:`D_i` as the
        censoring time. Survival data consist of the event indicator, :math:`\delta_i=1(X_i\leq D_i)`
        (argument ``event``) and the time-to-event or censoring, :math:`T_i = \min(\{ X_i,D_i \})`
        (argument ``time``).

        The log hazard function for the Cox proportional hazards model has the form:

        .. math::

            \log \lambda_i (t) = \log \lambda_{0}(t) + \log \theta_i

        where :math:`\log \theta_i` is the log relative hazard (argument ``log_hz``).

        **No ties in event time.**
        If the set :math:`\{T_i: \delta_i = 1\}_{i = 1, \cdots, N}` represent unique event times (i.e., no ties),
        the standard Cox partial likelihood can be used :cite:p:`Cox1972`. Let :math:`\tau_1 < \tau_2 < \cdots < \tau_N`
        be the ordered times and let  :math:`R(\tau_i) = \{ j: \tau_j \geq \tau_i\}`
        be the risk set at :math:`\tau_i`. The partial log likelihood is defined as:

        .. math::

            pll = \sum_{i: \: \delta_i = 1} \left(\log \theta_i - \log\left(\sum_{j \in R(\tau_i)} \theta_j \right) \right)

        **Ties in event time handled with Breslow's method.**
        Breslow's method :cite:p:`Breslow1975` describes the approach in which the procedure described above is used unmodified,
        even when ties are present. If two subjects A and B have the same event time, subject A will be at risk for the
        event that happened to B, and B will be at risk for the event that happened to A.
        Let :math:`\xi_1 < \xi_2 < \cdots` denote the unique ordered times (i.e., unique :math:`\tau_i`). Let :math:`H_k` be the set of
        subjects that have an event at time :math:`\xi_k` such that :math:`H_k = \{i: \tau_i = \xi_k, \delta_i = 1\}`, and let :math:`m_k`
        be the number of subjects that have an event at time :math:`\xi_k` such that :math:`m_k = |H_k|`.

        .. math::

            pll = \sum_{k} \left( {\sum_{i\in H_{k}}\log \theta_i} - m_k \: \log\left(\sum_{j \in R(\tau_k)} \theta_j \right) \right)


        **Ties in event time handled with Efron's method.**
        An alternative approach that is considered to give better results is the Efron's method :cite:p:`Efron1977`.
        As a compromise between the Cox's and Breslow's method, Efron suggested to use the average
        risk among the subjects that have an event at time :math:`\xi_k`:

        .. math::

            \bar{\theta}_{k} = {\frac {1}{m_{k}}}\sum_{i\in H_{k}}\theta_i

        Efron approximation of the partial log likelihood is defined by

        .. math::

            pll = \sum_{k} \left( {\sum_{i\in H_{k}}\log \theta_i} - \sum_{r =0}^{m_{k}-1} \log\left(\sum_{j \in R(\xi_k)}\theta_j-r\:\bar{\theta}_{j}\right)\right)


    Examples:
        >>> log_hz = torch.tensor([0.1, 0.2, 0.3, 0.4, 0.5])
        >>> event = torch.tensor([1, 0, 1, 0, 1], dtype=torch.bool)
        >>> time = torch.tensor([1., 2., 3., 4., 5.])
        >>> neg_partial_log_likelihood(log_hz, event, time) # default, mean of log likelihoods across patients
        tensor(1.0071)
        >>> neg_partial_log_likelihood(log_hz, event, time, reduction = 'sum') # sum of log likelihoods across patients
        tensor(3.0214)
        >>> time = torch.tensor([1., 2., 2., 4., 5.])  # Dealing with ties (default: Efron)
        >>> neg_partial_log_likelihood(log_hz, event, time, ties_method = "efron")
        tensor(1.0873)
        >>> neg_partial_log_likelihood(log_hz, event, time, ties_method = "breslow")  # Dealing with ties (Breslow)
        tensor(1.0873)

    References:

        .. bibliography::
            :filter: False

            Cox1972
            Breslow1975
            Efron1977

    """

    if checks:
        validate_loss(log_hz, event, time, model_type="cox")

    if any([event.sum().item() == 0, len(log_hz.size()) == 0]):
        warnings.warn("No events OR single sample. Returning zero loss for the batch")
        return torch.tensor(0.0, requires_grad=True)

    # sort data by time-to-event or censoring
    time_sorted, idx = torch.sort(time)
    log_hz_sorted = log_hz[idx]
    event_sorted = event[idx]
    time_unique = torch.unique(time_sorted)  # time-to-event or censoring without ties

    if len(time_unique) == len(time_sorted):
        # if not ties, use traditional cox partial likelihood
        pll = _partial_likelihood_cox(log_hz_sorted, event_sorted)
    else:
        # add warning about ties
        warnings.warn(
            f"Ties in event time detected; using {ties_method}'s method to handle ties."
        )
        # if ties, use either efron or breslow approximation of partial likelihood
        if ties_method == "efron":
            pll = _partial_likelihood_efron(
                log_hz_sorted,
                event_sorted,
                time_sorted,
                time_unique,
            )
        elif ties_method == "breslow":
            pll = _partial_likelihood_breslow(log_hz_sorted, event_sorted, time_sorted)
        else:
            raise ValueError(
                f'Ties method {ties_method} should be one of ["efron", "breslow"]'
            )

    # Negative partial log likelihood
    pll = torch.neg(pll)
    if reduction.lower() == "mean":
        loss = pll.nanmean()
    elif reduction.lower() == "sum":
        loss = pll.sum()
    else:
        raise (
            ValueError(
                f"Reduction {reduction} is not implemented yet, should be one of ['mean', 'sum']."
            )
        )
    return loss


if __name__ == "__main__":
    import doctest

    # Run doctest
    results = doctest.testmod()
    if results.failed == 0:
        print("All tests passed.")
    else:
        print("Some doctests failed.")
        sys.exit(1)



import json
import os
import unittest

import numpy as np
import torch

from torchsurv.loss.cox import neg_partial_log_likelihood as cox
from torchsurv.tools.validate_data import validate_loss

# Load the benchmark cox log likelihoods from R
with open("tests/benchmark_data/benchmark_cox.json") as file:
    benchmark_cox_logliks = json.load(file)

# set seed for reproducibility
torch.manual_seed(42)
np.random.seed(42)


class TestCoxSurvivalLoss(unittest.TestCase):
    """
    List of packages compared
        - survival (R)
    """

    # random data and parameters
    N = 32
    log_hz = torch.randn(N)
    event = torch.randint(low=0, high=2, size=(N,)).bool()
    time = torch.randint(low=1, high=100, size=(N,))

    def test_y_tensor(self):
        event_np_array = np.random.randint(0, 1 + 1, size=(self.N,), dtype="bool")
        with self.assertRaises((RuntimeError, TypeError)):
            cox(self.log_hz, event_np_array, self.time)

    def test_t_tensor(self):
        time_np_array = np.random.randint(0, 100, size=(self.N,))
        with self.assertRaises((RuntimeError, TypeError)):
            cox(self.log_hz, self.event, time_np_array)

    def test_log_hz_tensor(self):
        log_hz_np_array = np.random.randn(
            self.N,
        )
        with self.assertRaises((RuntimeError, TypeError)):
            cox(log_hz_np_array, self.event, self.time)

    def test_len_data(self):
        time_wrong_len = torch.randint(low=1, high=100, size=(self.N + 1,))
        with self.assertRaises(TypeError):
            validate_loss(self.log_hz, self.event, time_wrong_len)

    def test_positive_t(self):
        time_negative = torch.randint(low=-100, high=100, size=(self.N,))
        with self.assertRaises(TypeError):
            validate_loss(self.log_hz, self.event, time_negative)

    def test_boolean_y(self):
        event_non_boolean = torch.randint(low=0, high=3, size=(self.N,))
        with self.assertRaises(TypeError):
            validate_loss(self.log_hz, event_non_boolean, self.time)

    def test_log_likelihood_without_ties(self):
        """test cox partial log likelihood without ties on lung and gbsg datasets"""
        for benchmark_cox_loglik in benchmark_cox_logliks:
            if benchmark_cox_loglik["no_ties"][0] == True:
                log_lik = -cox(
                    torch.tensor(
                        benchmark_cox_loglik["log_hazard"], dtype=torch.float32
                    ).squeeze(0),
                    torch.tensor(benchmark_cox_loglik["status"]).bool(),
                    torch.tensor(benchmark_cox_loglik["time"], dtype=torch.float32),
                    reduction="sum",
                )
                log_lik_survival = benchmark_cox_loglik["log_likelihood"]

                self.assertTrue(
                    np.allclose(
                        log_lik.numpy(),
                        np.array(log_lik_survival),
                        rtol=1e-5,
                        atol=1e-8,
                    )
                )

    def test_log_likelihood_with_ties(self):
        """test Efron and Breslow's approximation of cox partial log likelihood with ties on lung and gbsg data"""
        for benchmark_cox_loglik in benchmark_cox_logliks:
            if benchmark_cox_loglik["no_ties"][0] == False:
                # efron approximation of partial log likelihood
                log_lik_efron = -cox(
                    torch.tensor(
                        benchmark_cox_loglik["log_hazard_efron"], dtype=torch.float32
                    ).squeeze(0),
                    torch.tensor(benchmark_cox_loglik["status"]).bool(),
                    torch.tensor(benchmark_cox_loglik["time"], dtype=torch.float32),
                    ties_method="efron",
                    reduction="sum",
                )
                log_lik_efron_survival = benchmark_cox_loglik["log_likelihood_efron"]

                # breslow approximation of partial log likelihood
                log_lik_breslow = -cox(
                    torch.tensor(
                        benchmark_cox_loglik["log_hazard_breslow"], dtype=torch.float32
                    ).squeeze(0),
                    torch.tensor(benchmark_cox_loglik["status"]).bool(),
                    torch.tensor(benchmark_cox_loglik["time"], dtype=torch.float32),
                    ties_method="breslow",
                    reduction="sum",
                )
                log_lik_breslow_survival = benchmark_cox_loglik[
                    "log_likelihood_breslow"
                ]

                self.assertTrue(
                    np.allclose(
                        log_lik_efron.numpy(),
                        np.array(log_lik_efron_survival),
                        rtol=1e-5,
                        atol=1e-8,
                    )
                )

                self.assertTrue(
                    np.allclose(
                        log_lik_breslow.numpy(),
                        np.array(log_lik_breslow_survival),
                        rtol=1e-5,
                        atol=1e-8,
                    )
                )


if __name__ == "__main__":
    unittest.main()



import warnings
from typing import Optional

import torch

from torchsurv.stats import kaplan_meier
from torchsurv.tools.validate_data import validate_survival_data

__all__ = [
    "get_ipcw",
    "_inverse_censoring_dist",
]


# pylint: disable=anomalous-backslash-in-string
def get_ipcw(
    event: torch.Tensor,
    time: torch.Tensor,
    new_time: Optional[torch.Tensor] = None,
    checks: bool = True,
) -> torch.Tensor:
    r"""Calculate the inverse probability censoring weights (IPCW).

    Args:
        event (torch.Tensor, boolean):
            Event indicator of size n_samples (= True if event occurred).
        time (torch.Tensor, float):
            Time-to-event or censoring of size n_samples.
        new_time (torch.Tensor, float, optional):
            New time at which to evaluate the IPCW.
            Defaults to ``time``.
        checks (bool):
            Whether to perform input format checks.
            Enabling checks can help catch potential issues in the input data.
            Defaults to True.
    Returns:
        torch.Tensor: IPCW evaluated at ``new_time``.

    Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 5
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> time = torch.randint(low=1, high=100, size=(n,)).float()
            >>> new_time = torch.randint(low=1, high=100, size=(n*2,))
            >>> get_ipcw(event, time) # ipcw evaluated at time
            tensor([1.8750, 1.2500, 3.7500, 0.0000, 1.2500])
            >>> get_ipcw(event, time, new_time) # ipcw evaluated at new_time
            tensor([1.8750, 1.8750, 3.7500, 3.7500, 0.0000, 1.2500, 0.0000, 1.2500, 1.2500,
                    1.2500])

    Note:
        The inverse probability of censoring weight at time :math:`t` is defined by

        .. math::
            \omega(t) = 1 / \hat{D}(t),

        where :math:`\hat{D}(t)` is the Kaplan-Meier estimate of the censoring distribution, :math:`P(D>t)`.


    """

    if checks:
        validate_survival_data(event, time)

    # time on which to evaluate IPCW
    if new_time is None:  # if none, return ipcw of same size as time
        new_time = time

    # fit KM censoring estimator
    km = kaplan_meier.KaplanMeierEstimator()
    km(event, time, censoring_dist=True)

    # predict censoring distribution at time
    ct = km.predict(new_time)

    # calculate ipcw
    mask = ct > 0
    ipcw = torch.zeros_like(new_time, dtype=time.dtype)
    ipcw[mask] = _inverse_censoring_dist(ct[mask])

    return ipcw


def _inverse_censoring_dist(ct: torch.Tensor) -> torch.Tensor:
    """Compute inverse of the censoring distribution.

    Args:
        ct (torch.Tensor):
            Estimated censoring distribution.
            Must be strictly positive.

    Returns:
        torch.Tensor: 1/ct if ct > 0, else 1.

    Examples:
        >>> _ = torch.manual_seed(42)
        >>> ct = torch.randn((4,))
        >>> _inverse_censoring_dist(ct)
        tensor([2.9701, 7.7634, 4.2651, 4.3415])

    """
    if torch.any(ct == 0.0):
        zero_indices = torch.nonzero(ct.eq(0.0)).squeeze()
        zero_indices_list = zero_indices.tolist()  # Explicitly convert to list
        warnings.warn(
            f"Censoring distribution zero at time points: {zero_indices_list}. Returning ones as weight"
        )
    weight = 1.0 / ct
    weight = torch.ones_like(ct) / ct
    return weight


if __name__ == "__main__":
    import doctest

    # Run doctest
    results = doctest.testmod()
    if results.failed == 0:
        print("All tests passed.")



import copy
import sys
from typing import Optional

import torch
from scipy import stats
from torchmetrics import regression

from torchsurv.stats import kaplan_meier
from torchsurv.tools.validate_data import (
    validate_log_shape,
    validate_new_time,
    validate_survival_data,
)

__all__ = ["Auc"]


class Auc:
    """Area Under the Curve class for survival models."""

    def __init__(self, checks: bool = True, tied_tol: float = 1e-8):
        """Initialize an Auc for survival class model evaluation.

        Args:
            tied_tol (float):
                Tolerance for tied risk scores.
                Defaults to 1e-8.
            checks (bool):
                Whether to perform input format checks.
                Enabling checks can help catch potential issues in the input data.
                Defaults to True.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 10
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.randn((n,))
            >>> auc = Auc()
            >>> auc(estimate, event, time) # default: auc cumulative/dynamic
            tensor([0.7500, 0.4286, 0.3333])
            >>> auc.integral()
            tensor(0.5040)
            >>> auc.confidence_interval() # default: Blanche, two_sided
            tensor([[0.4213, 0.0000, 0.0000],
                    [1.0000, 0.9358, 0.7289]])
            >>> auc.p_value()
            tensor([0.1360, 0.7826, 0.4089])
        """
        self.tied_tol = tied_tol
        self.checks = checks

        # init instate variables
        self.order_time = None
        self.time = None
        self.event = None
        self.weight = None
        self.new_time = None
        self.weight_new_time = None
        self.estimate = None
        self.is_case = None
        self.is_control = None
        self.auc_type = None
        self.auc = None

    def __call__(
        self,
        estimate: torch.Tensor,
        event: torch.Tensor,
        time: torch.Tensor,
        auc_type: str = "cumulative",
        weight: Optional[torch.Tensor] = None,
        new_time: Optional[torch.Tensor] = None,
        weight_new_time: Optional[torch.Tensor] = None,
        instate: bool = True,
    ) -> torch.Tensor:
        r"""Compute the time-dependent Area Under the Receiver Operating Characteristic Curve (AUC).

        The AUC at time :math:`t` is the probability that a model correctly predicts
        which of two comparable samples will experience an event by time :math:`t` based on their
        estimated risk scores.
        The AUC is particularly useful for evaluating time-dependent predictions (e.g., 10-year mortality).
        It is recommended to use AUC instead of the concordance index for such time-dependent predictions, as
        AUC is proper in this context, while the concordance index is not :cite:p:`Blanche2018`.

        Args:
            estimate (torch.Tensor):
                Estimated risk of event occurrence (i.e., risk score).
                Can be of shape = (n_samples,) if subject-specific risk score is time-independent,
                of shape = (n_samples, n_samples) if subject-specific risk score is evaluated at ``time``,
                or of shape = (n_samples, n_times) if subject-specific risk score is evaluated at ``new_time``.
            event (torch.Tensor, boolean):
                Event indicator of size n_samples (= True if event occurred).
            time (torch.Tensor, float):
                Time-to-event or censoring of size n_samples.
            auc_type (str, optional):
                AUC type. Defaults to "cumulative".
                Must be one of the following: "cumulative" for cumulative/dynamic, "incident" for incident/dynamic.
            weight (torch.Tensor, optional):
                Optional sample weight evaluated at ``time`` of size n_samples.
                Defaults to 1.
            new_time (torch.Tensor, optional):
                Time points at which to evaluate the AUC of size n_times. Values must be within the range of follow-up in ``time``.
                Defaults to the event times excluding maximum (because number of controls for t > max(time) is 0).
            weight_new_time (torch.Tensor):
                Optional sample weight evaluated at ``new_time`` of size n_times.
                Defaults to 1.

        Returns:
            torch.Tensor: AUC evaluated at ``new_time``.

        Note:
            The function evaluates either the cumulative/dynamic (C/D) or the incident/dynamic (I/D) AUC (argument ``auc_type``)
            at time :math:`t \in \{t_1, \cdots, t_K\}` (argument ``new_time``).

            For each subject :math:`i \in \{1, \cdots, N\}`, denote :math:`X_i` as the survival time and :math:`D_i` as the
            censoring time. Survival data consist of the event indicator, :math:`\delta_i=(X_i\leq D_i)`
            (argument ``event``) and the time-to-event or censoring, :math:`T_i = \min(\{ X_i,D_i \})`
            (argument ``time``).

            The risk score measures the risk (or a proxy thereof) that a subject has an event.
            The function accepts time-dependent risk score or time-independent risk score. The time-dependent risk score
            of subject :math:`i` is specified through a function :math:`q_i: [0, \infty) \rightarrow \mathbb{R}`.
            The time-independent risk score of subject :math:`i` is specified by a constant :math:`q_i`.
            The argument ``estimate`` is the estimated risk score.
            For time-dependent risk score: if ``new_time`` is specified, the argument ``estimate`` should be of shape = (N,K)
            (:math:`(i,k)` th element is :math:`\hat{q}_i(t_k)`);
            if ``new_time`` is not specified, the argument ``estimate`` should be of shape = (N,N)
            (:math:`(i,j)` th element is :math:`\hat{q}_i(T_j)`) .
            For time-independent risk score, the argument ``estimate`` should be of length
            N (:math:`i` th element is :math:`\hat{q}_i`).

            The AUC C/D and AUC I/D evaluated at time :math:`t` are defined by

            .. math::
                \text{AUC}^{C/D}(t) = p(q_i(t) > q_j(t) \: | \: X_i \leq t, X_j > t) \\
                \text{AUC}^{I/D}(t) = p(q_i(t) > q_j(t) \: | \: X_i = t, X_j > t).


            The default estimators of the AUC C/D and AUC I/D at time :math:`t` :cite:p:`Blanche2013` returned by the function are

            .. math::

                \hat{\text{AUC}}^{C/D}(t) = \frac{\sum_i \sum_j \delta_i \: I(T_i \leq t, T_j > t) I(\hat{q}_i(t) > \hat{q}_j(t))}{\sum_i \delta_i \: I(T_i \leq t) \sum_j I(T_j > t)} \\
                \hat{\text{AUC}}^{I/D}(t) = \frac{\sum_i \sum_j \delta_i \: I(T_i = t, T_j > t) I(\hat{q}_i(t) > \hat{q}_j(t))}{\sum_i \delta_i \: I(T_i = t) \sum_j I(T_j > t)}.

            These estimators are considered naive because, when the event times are censored, all subjects censored
            before time point :math:`t` are ignored. Additionally, the naive estimators
            converge to an association measure that involves the censoring distribution.
            To address this shortcoming, :cite:t:`Uno2007` proposed to employ the
            inverse probability weighting technique. In this context, each subject included at time
            :math:`t` is weighted by the inverse probability of censoring :math:`\omega(t) = 1 / \hat{D}(t)`, where
            :math:`\hat{D}(t)` is the Kaplan-Meier estimate of the censoring distribution, :math:`P(D>t)`.
            The censoring-adjusted AUC C/D estimate at time :math:`t` is

            .. math::

                \hat{\text{AUC}}^{C/D}(t) = \frac{\sum_i \sum_j \delta_i \: \omega(T_i) \: I(T_i \leq t, T_j > t) I(\hat{q}_i(t) > \hat{q}_j(t))}{\sum_i \delta_i \: \omega(T_i) \: I(T_i \leq t)  \sum_j I(T_j > t)} \\

            Note that the censoring-adjusted AUC I/D estimate is the same as the "naive" estimate because the weights are all equal to :math:`\omega(t)`.

            The censoring-adjusted AUC C/D estimate can be obtained by specifying the argument
            ``weight``, the weights evaluated at each ``time`` (:math:`\omega(T_1), \cdots, \omega(T_N)`).
            If ``new_time`` is specified, the argument  ``weight_new_time``
            should also be specified accordingly, the weights evaluated at each ``new_time``
            (:math:`\omega(t_1), \cdots, \omega(t_K)`). The latter is required to compute the standard error of the AUC.
            In the context of train/test split, the weights should be derived from the censoring distribution estimated in the training data.
            Specifically, the censoring distribution is estimated using the training set and then evaluated at the subject time within the test set.

        Examples:
            >>> from torchsurv.stats.ipcw import get_ipcw
            >>> _ = torch.manual_seed(42)
            >>> n = 20
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.randn((n,))
            >>> auc = Auc()
            >>> auc(estimate, event, time) # default: naive auc c/d
            tensor([0.9474, 0.5556, 0.5294, 0.6429, 0.5846, 0.6389, 0.5844, 0.5139, 0.4028,
                    0.5400, 0.4545, 0.7500])
            >>> auc(estimate, event, time, auc_type = "incident") # naive auc i/d
            tensor([0.9474, 0.1667, 0.4706, 0.9286, 0.3846, 0.8333, 0.3636, 0.2222, 0.0000,
                    0.8000, 0.5000, 1.0000])
            >>> ipcw = get_ipcw(event, time) # ipcw weight at time
            >>> auc(estimate, event, time, weight = ipcw) # Uno's auc c/d
            tensor([0.9474, 0.5556, 0.5294, 0.6521, 0.5881, 0.6441, 0.5865, 0.5099, 0.3929,
                    0.5422, 0.4534, 0.7996])
            >>> new_time = torch.unique(torch.randint(low=100, high=150, size=(n,)).float()) # new time at which to evaluate auc
            >>> ipcw_new_time = get_ipcw(event, time, new_time) # ipcw at new_time
            >>> auc(estimate, event, time, new_time = new_time, weight = ipcw, weight_new_time = ipcw_new_time) # Uno's auc c/d at new_time
            tensor([0.5333, 0.5333, 0.5333, 0.5333, 0.6521, 0.6521, 0.5881, 0.5881, 0.5865,
                    0.5865, 0.5865, 0.5865, 0.5865, 0.6018, 0.6018, 0.5099])

        References:

            .. bibliography::
                :filter: False

                Blanche2018
                Blanche2013
                Uno2007
        """

        # mandatory input format checks
        self._validate_auc_inputs(
            estimate, time, auc_type, new_time, weight, weight_new_time
        )

        # update inputs as required
        estimate, new_time, weight, weight_new_time = self._update_auc_new_time(
            estimate, event, time, new_time, weight, weight_new_time
        )
        estimate = self._update_auc_estimate(estimate, new_time)
        weight, weight_new_time = self._update_auc_weight(
            time, new_time, weight, weight_new_time
        )

        # further input format checks
        if self.checks:
            validate_survival_data(event, time)
            validate_new_time(new_time, time)
            validate_log_shape(estimate)

        # sample size and length of time
        n_samples, n_times = estimate.shape[0], new_time.shape[0]

        # Expand arrays to (n_samples, n_times) shape
        time_long = time.unsqueeze(1).expand((n_samples, n_times))
        event_long = event.unsqueeze(1).expand((n_samples, n_times))
        weight_long = weight.unsqueeze(1).expand((n_samples, n_times))
        new_time_long = new_time.unsqueeze(0).expand(
            n_samples, n_times
        )  # Size n_times, unsqueeze(0) instead

        # sort each time point (columns) by risk score (descending)
        index = torch.argsort(-estimate, dim=0)
        time_long, event_long, estimate, weight_long = map(
            lambda x: torch.gather(x, 0, index),
            [time_long, event_long, estimate, weight_long],
        )

        # find case and control
        if auc_type == "incident":
            is_case = (time_long == new_time_long) & event_long
        elif auc_type == "cumulative":
            is_case = (time_long <= new_time_long) & event_long
        is_control = time_long > new_time_long
        n_controls = is_control.sum(axis=0)

        # estimate censoring adjusted true positive rate and false positive rate
        cumsum_tp = torch.cumsum(is_case * weight_long, axis=0)
        cumsum_fp = torch.cumsum(is_control, axis=0)
        true_pos = cumsum_tp / cumsum_tp[-1]
        false_pos = cumsum_fp / n_controls

        # prepend row of infinity values
        estimate_diff = torch.cat(
            (torch.full((1, n_times), float("inf")), estimate), dim=0
        )
        is_tied = torch.abs(torch.diff(estimate_diff, dim=0)) <= self.tied_tol

        # initialize empty tensor to store auc
        auc = torch.zeros_like(new_time)

        # iterate over time
        for i in range(n_times):
            # Extract relevant columns for the current time point
            tp, fp, mask = true_pos[:, i], false_pos[:, i], is_tied[:, i]

            # Create a boolean mask with False at the indices where ties occur
            mask_no_ties = torch.ones(tp.numel(), dtype=torch.bool)
            mask_no_ties[torch.nonzero(mask) - 1] = False

            # Extract values without ties using the boolean mask
            tp_no_ties = tp[mask_no_ties]
            fp_no_ties = fp[mask_no_ties]

            # Add an extra threshold position
            # to make sure that the curve starts at (0, 0)
            tp_no_ties = torch.cat([torch.tensor([0]), tp_no_ties])
            fp_no_ties = torch.cat([torch.tensor([0]), fp_no_ties])

            # Calculate AUC using trapezoidal rule for numerical integration
            auc[i] = torch.trapz(tp_no_ties, fp_no_ties)

        # Create/overwrite internal attributes states
        if instate:
            index_rev = torch.argsort(index, dim=0)
            estimate, is_case, is_control = map(
                lambda x: torch.gather(x, 0, index_rev),
                [estimate, is_case, is_control],
            )

            # sort all objects by time
            self.order_time = torch.argsort(time, dim=0)
            self.time = time[self.order_time]
            self.event = event[self.order_time]
            self.weight = weight[self.order_time]
            self.new_time = new_time
            self.weight_new_time = weight_new_time
            self.estimate = torch.index_select(estimate, 0, self.order_time)
            self.is_case = torch.index_select(is_case, 0, self.order_time)
            self.is_control = torch.index_select(is_control, 0, self.order_time)
            self.auc_type = auc_type
            self.auc = auc

        return auc

    def integral(self, tmax: Optional[torch.Tensor] = None):
        """Compute the integral of the time-dependent AUC.

        Args:
            tmax (torch.Tensor, optional):
                A number specifying the upper limit of the time range to compute the AUC integral.
                Defaults to ``new_time[-1]`` for cumulative/dynamic AUC and ``new_time[-1]-1`` for incident/dynamic AUC.

        Returns:
            torch.Tensor: Integral of AUC over [0-``tmax``].

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 10
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.randn((n,))
            >>> auc = Auc()
            >>> auc(estimate, event, time, auc_type = "incident")
            tensor([0.7500, 0.1429, 0.1667])
            >>> auc.integral() # integral of the auc incident/dynamic
            tensor(0.4667)

        Notes:
            In case ``auc_type`` = "cumulative" (cumulative/dynamic AUC), the values of AUC are weighted by
            the estimated event density. In case ``auc_type`` = "incident"
            (incident/dynamic AUC), the values of AUC are weighted by 2 times the product of the estimated
            event density and the estimated survival function :cite:p:`Heagerty2005`.
            The estimated survival
            function is the Kaplan-Meier estimate. The estimated event density
            is obtained from the discrete incremental changes of the estimated survival function.

        References:

            .. bibliography::
                :filter: False

                Heagerty2005
        """
        # Only one time step available
        if len(self.new_time) == 1:
            auc = self.auc[0]
        else:
            # handle cases where tmax is not specified
            if tmax is None:
                tmax = (
                    self.new_time[-1]
                    if self.auc_type == "cumulative"
                    else self.new_time[-1] - 1
                )

            # estimate of survival distribution
            km = kaplan_meier.KaplanMeierEstimator()
            km(self.event, self.time)
            survival = km.predict(self.new_time)

            # integral of auc
            if self.auc_type == "cumulative":
                auc = self._integrate_cumulative(survival, tmax)
            else:
                auc = self._integrate_incident(survival, tmax)
        return auc

    def confidence_interval(
        self,
        method: str = "blanche",
        alpha: float = 0.05,
        alternative: str = "two_sided",
        n_bootstraps: int = 999,
    ) -> torch.Tensor:
        """Compute the confidence interval of the AUC.

        This function calculates either the pointwise confidence interval or the bootstrap
        confidence interval for the AUC. The pointwise confidence interval is computed
        assuming that the AUC is normally distributed and using the standard error estimated with
        :cite:t:`Blanche2013b` method. The bootstrap confidence interval is constructed based on the
        distribution of bootstrap samples.

        Args:
            method (str):
                Method for computing confidence interval. Defaults to "blanche".
                Must be one of the following: "blanche", "bootstrap".
            alpha (float):
                Significance level. Defaults to 0.05.
            alternative (str):
                Alternative hypothesis. Defaults to "two_sided".
                Must be one of the following: "two_sided", "greater", "less".
            n_bootstraps (int):
                Number of bootstrap samples. Defaults to 999.
                Ignored if ``method`` is not "bootstrap".

        Returns:
            torch.Tensor([lower,upper]):
                Lower and upper bounds of the confidence interval.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 10
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.randn((n,))
            >>> auc = Auc()
            >>> auc(estimate, event, time)
            tensor([0.7500, 0.4286, 0.3333])
            >>> auc.confidence_interval() # Default: Blanche, two_sided
            tensor([[0.4213, 0.0000, 0.0000],
                    [1.0000, 0.9358, 0.7289]])
            >>> auc.confidence_interval(method = "bootstrap", alternative = "greater")
            tensor([[0.3750, 0.1667, 0.0833],
                    [1.0000, 1.0000, 1.0000]])

        References:

            .. bibliography::
                :filter: False

                Blanche2013b
        """

        assert (
            hasattr(self, "auc") and self.auc is not None
        ), "Error: Please calculate AUC using `Auc()` before calling `confidence_interval()`."

        if alternative not in ["less", "greater", "two_sided"]:
            raise ValueError(
                "'alternative' parameter must be one of ['less', 'greater', 'two_sided']."
            )

        if method == "blanche":
            conf_int = self._confidence_interval_blanche(alpha, alternative)
        elif method == "bootstrap":
            conf_int = self._confidence_interval_bootstrap(
                alpha, alternative, n_bootstraps
            )
        else:
            raise ValueError(
                "Method not implemented. Please choose either 'blanche' or 'bootstrap'."
            )
        return conf_int

    def p_value(
        self,
        method: str = "blanche",
        alternative: str = "two_sided",
        n_bootstraps: int = 999,
    ) -> torch.Tensor:
        """Perform a one-sample hypothesis test on the AUC.

        This function calculates either the pointwise p-value or the bootstrap p-value
        for testing the null hypothesis that the AUC is equal to 0.5.
        The pointwise p-value is computed assuming that the AUC is normally distributed
        and using the standard error estimated using :cite:t:`Blanche2013b` method.
        The bootstrap p-value is derived by permuting risk predictions to estimate
        the sampling distribution under the null hypothesis.

        Args:
            method (str):
                Method for computing p-value. Defaults to "blanche".
                Must be one of the following: "blanche", "bootstrap".
            alternative (str):
                Alternative hypothesis. Defaults to "two_sided".
                Must be one of the following: "two_sided" (AUC is not equal to 0.5),
                "greater" (AUC is greater than 0.5), "less" (AUC is less than 0.5).
            n_bootstraps (int):
                Number of bootstrap samples. Defaults to 999.
                Ignored if ``method`` is not "bootstrap".

        Returns:
            torch.Tensor: p-value of the statistical test.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 10
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> estimate = torch.randn((n,))
            >>> auc = Auc()
            >>> auc(estimate, event, time)
            tensor([0.7500, 0.4286, 0.3333])
            >>> auc.p_value() # Default: Blanche, two_sided
            tensor([0.1360, 0.7826, 0.4089])
            >>> auc.p_value(method = "bootstrap", alternative = "greater")
            tensor([0.2400, 0.5800, 0.7380])

        """

        assert (
            hasattr(self, "auc") and self.auc is not None
        ), "Error: Please calculate AUC using `Auc()` before calling `p_value()`."

        if alternative not in ["less", "greater", "two_sided"]:
            raise ValueError(
                "'alternative' parameter must be one of ['less', 'greater', 'two_sided']."
            )

        if method == "blanche":
            pvalue = self._p_value_blanche(alternative)
        elif method == "bootstrap":
            pvalue = self._p_value_bootstrap(alternative, n_bootstraps)
        else:
            raise ValueError(
                "Method not implemented. Please choose either 'blanche' or 'bootstrap'."
            )
        return pvalue

    def compare(
        self, other, method: str = "blanche", n_bootstraps: int = 999
    ) -> torch.Tensor:
        """Compare two AUCs.

        This function compares two AUCs computed on the same data with different
        risk predictions. The statistical hypotheses are
        formulated as follows, null hypothesis: auc1 = auc2 and alternative
        hypothesis: auc1 > auc2.
        The statistical test is either a Student t-test for dependent samples or
        a two-sample bootstrap test. The Student t-test assumes that the AUC is normally distributed
        and uses the standard errors estimated with :cite:t:`Blanche2013b` method.

        Args:
            other (Auc):
                Another instance of the Auc class representing auc2.
            method (str):
                Statistical test used to perform the hypothesis test. Defaults to "blanche".
                Must be one of the following: "blanche", "bootstrap".
            n_bootstraps (int):
                Number of bootstrap samples. Defaults to 999.
                Ignored if ``method`` is not "bootstrap".

        Returns:
            torch.Tensor: p-value of the statistical test.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> n = 10
            >>> time = torch.randint(low=5, high=250, size=(n,)).float()
            >>> event = torch.randint(low=0, high=2, size=(n,)).bool()
            >>> auc1 = Auc()
            >>> auc1(torch.randn((n,)), event, time)
            tensor([0.7500, 0.4286, 0.3333])
            >>> auc2 = Auc()
            >>> auc2(torch.randn((n,)), event, time)
            tensor([0.0000, 0.1429, 0.0556])
            >>> auc1.compare(auc2) # default: Blanche
            tensor([0.0008, 0.2007, 0.1358])
            >>> auc1.compare(auc2, method = "bootstrap")
            tensor([0.0220, 0.1970, 0.1650])

        """

        assert (
            hasattr(self, "auc") and self.auc is not None
        ), "Error: Please calculate AUC using `Auc()` before calling `compare()`."

        # assert that the same data were used to compute the two auc
        if torch.any(self.event != other.event) or torch.any(self.time != other.time):
            raise ValueError(
                "Mismatched survival data: 'time' and 'event' should be the same for both AUC computations."
            )
        if torch.any(self.new_time != other.new_time):
            raise ValueError(
                "Mismatched evaluation times: 'new_time' should be the same for both AUC computations."
            )
        if self.auc_type != other.auc_type:
            raise ValueError(
                "Mismatched AUC types: 'auc_type' should be the same for both AUC computations."
            )

        if method == "blanche":
            pvalue = self._compare_blanche(other)
        elif method == "bootstrap":
            pvalue = self._compare_bootstrap(other, n_bootstraps)
        else:
            raise ValueError(
                f"Method {method} not implemented. Please choose either 'blanche' or 'bootstrap'."
            )
        return pvalue

    # pylint: disable=invalid-name
    def _integrate_incident(self, S: torch.Tensor, tmax: torch.Tensor) -> torch.Tensor:
        """Integrates the incident/dynamic AUC, int_t AUC(t) x w(t) dt
        where w(t) = 2*f(t)*S(t) and f(t) is the lifeline distribution,
        S(t) is the survival distribution estimated with the Kaplan
        Meier estimator.
        """

        # Find the index corresponding to tmax in times
        tmax_index = torch.sum(self.new_time <= tmax)

        # Initialize an array to store the density function f(t)
        f = torch.zeros_like(S)

        # Compute the density function f(t)
        f[0] = 1.0 - S[0]
        for i in range(1, len(S)):
            f[i] = S[i - 1] - S[i]

        # Initialize a variable to accumulate the weighted sum of f(t)*S(t)
        wT = 0.0

        # Accumulate the weighted sum up to maxI
        for i in range(tmax_index):
            wT += 2.0 * f[i] * S[i]

        # Initialize the integrated AUC
        i_auc = 0

        # Calculate the integrated AUC using the weight
        for i in range(tmax_index):
            if wT != 0.0:
                if f[i] > torch.finfo(torch.double).eps:
                    i_auc += self.auc[i] * (2.0 * f[i] * S[i]) / wT

        return i_auc

    # pylint: disable=invalid-name
    def _integrate_cumulative(
        self, S: torch.Tensor, tmax: torch.Tensor
    ) -> torch.Tensor:
        """Integrates the cumulative/dynamic AUC, int_t AUC(t) · f(t) dt
        where f(t) is the lifeline distribution estimated from the discrete
        incremental changes of the Kalpan-Meier estimate of the survival function.
        """

        # Find the index corresponding to tmax in times

        tmax_index = torch.sum(self.new_time <= tmax)

        # Initialize an array to store the density function f(t)
        f = torch.zeros_like(S)

        # Compute the density function f(t)
        f[0] = 1.0 - S[0]
        for i in range(1, len(S)):
            f[i] = S[i - 1] - S[i]

        # Initialize a variable to accumulate the weighted sum of f(t)
        wT = 0.0

        # Accumulate the weighted sum up to maxI
        for i in range(tmax_index):
            if f[i] > torch.finfo(torch.double).eps:
                wT += f[i]

        # Initialize the integrated AUC
        i_auc = 0

        # Calculate the integrated AUC using the weight
        for i in range(tmax_index):
            if wT != 0.0:
                if f[i] > torch.finfo(torch.double).eps:
                    i_auc += self.auc[i] * (f[i]) / wT

        return i_auc

    def _confidence_interval_blanche(
        self, alpha: float, alternative: str
    ) -> torch.Tensor:
        """Confidence interval of AUC assuming that the AUC
        is normally distributed and using
        standard error estimated with Blanche et al.'s method."""

        alpha = alpha / 2 if alternative == "two_sided" else alpha

        auc_se = self._auc_se()

        if torch.all(auc_se) >= 0:
            ci = (
                -torch.distributions.normal.Normal(0, 1).icdf(torch.tensor(alpha))
                * auc_se
            )
            lower = torch.max(torch.tensor(0.0), self.auc - ci)
            upper = torch.min(torch.tensor(1.0), self.auc + ci)

            if alternative == "less":
                lower = torch.zeros_like(lower)
            elif alternative == "greater":
                upper = torch.ones_like(upper)
        else:
            raise ValueError("The standard error of AUC must be a positive value.")

        return torch.stack([lower, upper], dim=0).squeeze()

    def _confidence_interval_bootstrap(
        self, alpha: float, alternative: str, n_bootstraps: int
    ) -> torch.Tensor:
        """Bootstrap confidence interval of the AUC using Efron percentile method.

        References:
            Efron, Bradley; Tibshirani, Robert J. (1993).
                An introduction to the bootstrap, New York: Chapman & Hall, software.
        """

        # auc bootstraps given bootstrap distribution
        auc_bootstrap = self._bootstrap_auc(
            metric="confidence_interval", n_bootstraps=n_bootstraps
        )

        # initialize tensor to store confidence intervals
        lower = torch.zeros_like(self.auc)
        upper = torch.zeros_like(self.auc)

        # iterate over time
        for index_t in range(len(self.auc)):
            # obtain confidence interval
            if alternative == "two_sided":
                lower[index_t], upper[index_t] = torch.quantile(
                    auc_bootstrap[:, index_t],
                    torch.tensor([alpha / 2, 1 - alpha / 2], device=self.auc.device),
                )
            elif alternative == "less":
                upper[index_t] = torch.quantile(
                    auc_bootstrap[:, index_t],
                    torch.tensor(1 - alpha, device=self.auc.device),
                )
                lower[index_t] = torch.tensor(0.0, device=self.auc.device)
            elif alternative == "greater":
                lower[index_t] = torch.quantile(
                    auc_bootstrap[:, index_t],
                    torch.tensor(alpha, device=self.auc.device),
                )
                upper[index_t] = torch.tensor(1.0, device=self.auc.device)

        return torch.stack([lower, upper], dim=0).squeeze()

    def _p_value_blanche(
        self, alternative: str, null_value: float = 0.5
    ) -> torch.Tensor:
        """p-value for a one-sample hypothesis test of the AUC
        assuming that the AUC is normally distributed and using standard error estimated
        with Blanche et al's method.
        """

        auc_se = self._auc_se()

        # get p-value
        if torch.all(auc_se >= 0.0):
            p = torch.distributions.normal.Normal(0, 1).cdf(
                (self.auc - null_value) / auc_se
            )
            if alternative == "two_sided":
                mask = self.auc >= torch.tensor(0.5)
                p[mask] = torch.tensor(1.0) - p[mask]
                p *= torch.tensor(2.0)
            elif alternative == "greater":
                p = torch.tensor(1.0) - p
        else:
            raise ValueError("The standard error of AUC must be a positive value.")

        return p

    def _p_value_bootstrap(self, alternative, n_bootstraps) -> torch.Tensor:
        """p-value for a one-sample hypothesis test of the AUC using
        permutation of risk prediction to estimate sampling distribution under the null
        hypothesis.
        """

        # auc bootstraps given null distribution auc = 0.5
        auc0 = self._bootstrap_auc(metric="p_value", n_bootstraps=n_bootstraps)

        # initialize empty tensor to store p-values
        p_values = torch.zeros_like(self.auc)

        # iterate over time
        for index_t, _ in enumerate(self.auc):
            # Derive p-value
            p = (
                torch.tensor(1.0) + torch.sum(auc0[:, index_t] <= self.auc[index_t])
            ) / torch.tensor(n_bootstraps + 1.0)
            if alternative == "two_sided":
                if self.auc[index_t] >= torch.tensor(0.5):
                    p = torch.tensor(1.0) - p
                p *= torch.tensor(2.0)
                p = torch.min(
                    torch.tensor(1.0, device=self.auc.device), p
                )  # in case very small bootstrap sample size is used
            elif alternative == "greater":
                p = torch.tensor(1.0) - p

            p_values[index_t] = p

        return p_values

    def _compare_blanche(self, other) -> torch.Tensor:
        """Student t-test for dependent samples given Blanche's standard error to
        compare two AUCs.
        """

        # sample size
        N = self.estimate.shape[0]

        # compute noether standard error
        auc1_se = self._auc_se()
        # pylint: disable=protected-access
        auc2_se = other._auc_se()

        # initialize empty vector to store p_values
        p_values = torch.zeros_like(self.auc)

        # iterate over time
        for index_t, _ in enumerate(self.auc):
            # compute spearman correlation between risk prediction
            corr = regression.SpearmanCorrCoef()(
                self.estimate[:, index_t], other.estimate[:, index_t]
            )

            # check for perfect positive monotonic relationship between two variables
            if 1 - torch.abs(corr) < 1e-15:
                p_values[index_t] = 1.0
            else:
                # compute t-stat
                t_stat = (self.auc[index_t] - other.auc[index_t]) / torch.sqrt(
                    auc1_se[index_t] ** 2
                    + auc2_se[index_t] ** 2
                    - 2 * corr * auc1_se[index_t] * auc2_se[index_t]
                )

                # p-value
                p_values[index_t] = torch.tensor(
                    1
                    - stats.t.cdf(
                        t_stat, df=N - 1
                    ),  # student-t cdf not available on torch
                    dtype=self.auc.dtype,
                    device=self.auc.device,
                )

        return p_values

    def _compare_bootstrap(self, other, n_bootstraps) -> torch.Tensor:
        """Bootstrap two-sample test to compare two AUCs."""

        # auc bootstraps given null hypothesis that auc1 and
        # auc2 come from the same distribution
        auc1_null = self._bootstrap_auc(
            metric="compare", other=other, n_bootstraps=n_bootstraps
        )
        auc2_null = self._bootstrap_auc(
            metric="compare", other=other, n_bootstraps=n_bootstraps
        )

        # Bootstrapped test statistics
        t_boot = auc1_null - auc2_null

        # observed test statistics
        t_obs = self.auc - other.auc

        # initialize empty tensor to store p-values
        p_values = torch.zeros_like(self.auc)

        # iterate over time
        for index_t in range(len(self.auc)):
            p_values[index_t] = 1 - (
                1 + torch.sum(t_boot[:, index_t] <= t_obs[index_t])
            ) / (n_bootstraps + 1)

        return p_values

    def _auc_se(self) -> torch.Tensor:
        """AUC's standard error estimated using Blanche et al's method."""

        # sample size and length of time
        n_samples, n_times = self.estimate.shape[0], self.new_time.shape[0]

        # survival distribution estimated with KM
        km = kaplan_meier.KaplanMeierEstimator()
        km(self.event, self.time)
        S = km.predict(self.new_time)

        # Expand arrays to (n_samples, n_times) shape
        time_long = torch.unsqueeze(self.time, dim=1).expand((n_samples, n_times))
        new_time_long = self.new_time.unsqueeze(0).expand(n_samples, n_times)
        survival_long = S.unsqueeze(0).expand(n_samples, n_times)

        # element (i, t) = I(T_i > t) / S(t)
        ipsw = (time_long >= new_time_long) / survival_long

        # element (i,k) = int_0^{T_i} M_Ck(t) / S dt, where M_Ck(t) = I(delta_k = 0, T_k <= t) - int_0^t dLambda_c(u)
        integral_m_div_s = self._integral_censoring_martingale_divided_survival()

        # element (i, t) = f_i1t = I(T_i <= t, delta_i = 1) * W(T_i) (if incident, condition for case is T_i = t)
        f = self.is_case * self.weight.unsqueeze(1).expand(-1, n_times)

        # element (t) = F_1t = 1/n * sum_i f_i1t
        F = 1 / n_samples * torch.sum(f, dim=0)

        # initialize empty tensor to store standard error of auc
        auc_se = torch.zeros_like(self.new_time)

        # iterate over time
        for index_t in range(n_times):
            # element(i,j) given t = h*_tij = I(T_i <= t, delta_i = 1)  * I(U_j > t) * (I(M_i > M_j) + 0.5 * I(M_i > M_j)) * W(T_i) * W(t)
            h_t = (
                self.is_case[:, index_t].unsqueeze(1) * self.is_control[:, index_t]
            ).float()
            h_t *= (
                self.estimate[:, index_t].unsqueeze(1)
                > self.estimate[:, index_t].unsqueeze(0)
            ).float() + 0.5 * (
                self.estimate[:, index_t].unsqueeze(1)
                == self.estimate[:, index_t].unsqueeze(0)
            ).float()
            h_t *= self.weight.unsqueeze(1)
            h_t *= self.weight_new_time[index_t]

            # h*_t = 1 / n^2 * sum_i sum_j h_tij
            H_t = (1 / (n_samples**2)) * torch.sum(h_t)

            # phi*(t), eq.bottom page 3 of Supplementary Material of Blanche et al. (2013)
            phi = self._phi_blanche(
                h_t,
                H_t,
                f[:, index_t],
                F[index_t],
                S[index_t],
                integral_m_div_s,
                ipsw[:, index_t],
                n_samples,
            )

            # sum of phi
            sum_phi = (
                torch.sum(
                    phi, axis=(1, 2)
                )  # sum_j sum_k phi_ijk = Les_sum_jk_a_i_fixe * n_samples in Blanche's code
                + torch.sum(
                    phi, axis=(0, 2)
                )  # sum_i sum_k phi_ijk = Les_sum_ik_a_j_fixe * n_samples in Blanche's code
                + torch.sum(
                    phi, axis=(0, 1)
                )  # sum_i sum_j phi_ijk = Les_sum_ij_a_k_fixe * n_samples in Blanche's code
            )

            # estimate of influence function
            hat_IF = 1 / (n_samples**2) * sum_phi

            # estimate of auc standard error
            auc_se[index_t] = torch.std(hat_IF) / (n_samples ** (1 / 2))

        return auc_se

    def _integral_censoring_martingale_divided_survival(self) -> torch.Tensor:
        """Compute the integral of the censoring martingale divided by the survival distribution."""

        # Number of samples
        n_samples = len(self.time)

        # check that time is ordered
        if torch.any(self.time != self.time[torch.argsort(self.time, dim=0)]):
            raise ValueError(
                "The 'time' values in `self.time` should be ordered in ascending order."
            )

        # find censoring events
        censoring = self.event == 0.0

        # Compute censoring hazard, denoted lambda_C in in Blanche et al's paper
        censoring_hazard = censoring / torch.arange(n_samples, 0, -1)

        # cumulative censoring hazard, denoted Lambda_C in Blanche et al's paper
        cumsum_censoring_hazard = torch.cumsum(censoring_hazard, dim=0)

        # censoring martingale, denoted M_C in in Blanche et al's paper
        censoring_martingale = torch.zeros((n_samples, n_samples), dtype=torch.float32)
        for i in range(n_samples):
            censoring_martingale[:, i] = censoring[i].float() * (
                self.time[i] <= self.time
            ).float() - torch.cat(
                (
                    cumsum_censoring_hazard[0:i],
                    torch.full((n_samples - i,), cumsum_censoring_hazard[i]),
                )
            )

        # derivative of censoring martingal, denoted d M_C in Blanche et al's paper
        d_censoring_martingale = torch.cat(
            (
                censoring_martingale[0, :].view(1, -1),
                censoring_martingale[1:, :] - censoring_martingale[:-1, :],
            )
        )

        def divide_by_empirical_survival(v):
            return v / torch.cat(
                (torch.tensor([1.0]), 1 - torch.arange(1, len(v)) / len(v))
            )

        d_censoring_martingale_div_s = torch.stack(
            [
                divide_by_empirical_survival(d_censoring_martingale[:, i])
                for i in range(n_samples)
            ],
            dim=1,
        )

        return torch.cumsum(d_censoring_martingale_div_s, dim=0)

    # pylint: disable=invalid-name
    @staticmethod
    def _phi_blanche(
        h_t: torch.Tensor,
        H_t: torch.Tensor,
        f_t: torch.Tensor,
        F_t: torch.Tensor,
        S_t: torch.Tensor,
        integral_m_div_s: torch.Tensor,
        ipsw_t: torch.Tensor,
        n_samples: int,
    ) -> torch.Tensor:
        """Compute the three components of the influence function."""

        phi = torch.zeros(
            (n_samples, n_samples, n_samples),
            dtype=h_t.dtype,
            device=h_t.device,
        )
        for i in range(n_samples):
            _phi = 1 + integral_m_div_s[:, i]
            _phi *= f_t
            _phi -= F_t
            _phi /= F_t
            _phi += ipsw_t
            _phi *= -H_t
            _phi = _phi.unsqueeze(1).expand(-1, n_samples)
            _phi = _phi + (
                h_t * ((1 + integral_m_div_s[:, i]).unsqueeze(1).expand(-1, n_samples))
            )
            _phi /= S_t * F_t

            phi[:, :, i] = _phi

        return phi

    def _bootstrap_auc(
        self, metric: str, n_bootstraps: int, other=None
    ) -> torch.Tensor:
        """Compute bootstrap samples of the AUC.

        Args:
            metric (str): Must be one of the following: "p_value", "confidence_interval", "compare".
                If "p_value", computes bootstrap samples of the AUC given the sampling distribution
                under the null hypothesis (AUC = 0.5). If "confidence_interval", computes bootstrap
                samples of the AUC given the data distribution. If "compare", computes
                bootstrap samples of the AUC given the sampling distribution under the comparison test
                null hypothesis (auc1 = auc2).
            n_bootstraps (int): Number of bootstrap samples.
            other (optional, Auc):
                Another instance of the Auc class representing auc2.
                Only required for ``metric`` is equal to "compare".


        Returns:
            torch.Tensor: Bootstrap samples of AUC.
        """
        # Initiate empty list to store auc
        aucs = []

        # Get the bootstrap samples of auc
        for _ in range(n_bootstraps):
            if (
                metric == "p_value"
            ):  # bootstrap samples given null distribution (auc = 0.5)
                estimate = copy.deepcopy(self.estimate)
                estimate = estimate[
                    torch.randperm(estimate.shape[0]), :
                ]  # Shuffle estimate
                aucs.append(
                    self(
                        estimate,
                        self.event,
                        self.time,
                        self.auc_type,
                        self.weight,
                        self.new_time,
                        self.weight_new_time,
                        instate=False,
                    )
                )  # Run without saving internal state
            elif (
                metric == "confidence_interval"
            ):  # bootstrap samples given data distribution
                index = torch.randint(
                    low=0,
                    high=self.estimate.shape[0],
                    size=(self.estimate.shape[0] - 2,),
                )
                # auc is evaluated at new_time:
                # so need to keep the min and max time to ensure that new_time is within follow-up time
                index_max_time = torch.argmax(self.time).unsqueeze(0)
                index_min_time = torch.nonzero(
                    self.time == self.time[self.event].min()
                ).squeeze(1)
                index = torch.cat(
                    (
                        index,
                        index_max_time,
                        index_min_time,
                    ),
                    dim=0,
                )

                aucs.append(
                    self(
                        self.estimate[index, :],
                        self.event[index],
                        self.time[index],
                        self.auc_type,
                        self.weight[index],
                        self.new_time,
                        self.weight_new_time,
                        instate=False,
                    )
                )  # Run without saving internal state

            elif (
                metric == "compare"
            ):  # bootstrap samples given null distribution (auc1 = auc2)
                # index included in bootstrap samples
                index = torch.randint(
                    low=0,
                    high=self.estimate.shape[0] * 2,
                    size=(self.estimate.shape[0] - 2,),
                )

                # auc is evaluated at new_time: need to keep the indices corresponding to min and max time
                # to ensure that new_time is within follow-up time
                # with prob 0.5, take the index corresponding to max time from self and with prob 0.5 from other
                # same for min time
                index_max_time = torch.argmax(self.time).unsqueeze(0)
                index_max_time += (torch.rand(1) < 0.5) * self.estimate.shape[0]
                index_min_time = torch.nonzero(
                    self.time == self.time[self.event].min()
                ).squeeze(1)
                index_min_time += (torch.rand(1) < 0.5) * self.estimate.shape[0]
                index = torch.cat(
                    (index, index_min_time, index_max_time),
                    dim=0,
                )

                # with prob 0.5, take the weight_new_time from self and with prob 0.5 from other
                weight_new_time = (
                    self.weight_new_time
                    if torch.rand(1) < 0.5
                    else other.weight_new_time
                )

                aucs.append(
                    self(  # sample with replacement from pooled sample
                        torch.cat((self.estimate, other.estimate))[index, :],
                        torch.cat((self.event, other.event))[index],
                        torch.cat((self.time, other.time))[index],
                        self.auc_type,
                        torch.cat((self.weight, other.weight))[index],
                        self.new_time,
                        weight_new_time,
                        instate=False,
                    )
                )

        aucs = torch.stack(aucs)

        if torch.any(torch.isnan(aucs)):
            raise ValueError("The AUC computed using bootstrap should not be NaN.")

        return aucs

    @staticmethod
    def _find_torch_unique_indices(
        inverse_indices: torch.Tensor, counts: torch.Tensor
    ) -> torch.Tensor:
        """return unique_sorted_indices such that
        sorted_unique_tensor[inverse_indices] = original_tensor
        original_tensor[unique_sorted_indices] = sorted_unique_tensor

        Usage:
            _, inverse_indices, counts = torch.unique(
                    x, sorted=True, return_inverse=True, return_counts=True
            )
            sorted_unique_indices = Auc._find_torch_unique_indices(
                    inverse_indices, counts
            )
        """

        _, ind_sorted = torch.sort(inverse_indices, stable=True)
        cum_sum = counts.cumsum(0)
        cum_sum = torch.cat((torch.tensor([0]), cum_sum[:-1]))
        return ind_sorted[cum_sum]

    @staticmethod
    def _validate_auc_inputs(
        estimate, time, auc_type, new_time, weight, weight_new_time
    ):
        # check new_time and weight are provided, weight_new_time should be provided
        if all([new_time is not None, weight is not None, weight_new_time is None]):
            raise ValueError(
                "Please provide 'weight_new_time', the weight evaluated at 'new_time'."
            )

        # check if auc type is correct
        if auc_type not in ["cumulative", "incident"]:
            raise ValueError(
                "The 'auc_type' parameter must be either 'cumulative' or 'incident'."
            )

        # check if new_time are not specified and time-dependent estimate are not evaluated at time
        if new_time is None and estimate.ndim == 2 and estimate.shape[1] != 1:
            if len(time) != estimate.shape[1]:
                raise ValueError(
                    "Mismatched dimensions: The number of columns in 'estimate' does not match the length of 'time'. "
                    "Please provide the times at which 'estimate' is evaluated using the 'new_time' input."
                )

    @staticmethod
    def _update_auc_new_time(
        estimate: torch.Tensor,
        event: torch.Tensor,
        time: torch.Tensor,
        new_time: torch.Tensor,
        weight: torch.Tensor,
        weight_new_time: torch.Tensor,
    ) -> torch.Tensor:
        # update new time
        if (
            new_time is not None
        ):  # if new_time are specified: ensure it has the correct format
            # ensure that new_time are float
            if isinstance(new_time, int):
                new_time = torch.tensor([new_time]).float()

            # example if new_time is tensor(12), unsqueeze to tensor([12])
            if new_time.ndim == 0:
                new_time = new_time.unsqueeze(0)

        else:  # else: find new_time
            # if new_time are not specified, use unique event time
            mask = event & (time < torch.max(time))
            new_time, inverse_indices, counts = torch.unique(
                time[mask], sorted=True, return_inverse=True, return_counts=True
            )
            sorted_unique_indices = Auc._find_torch_unique_indices(
                inverse_indices, counts
            )

            # select weight corresponding at new time
            if weight is not None:
                weight_new_time = (weight[mask])[sorted_unique_indices]

            # for time-dependent estimate, select those corresponding to new time
            if estimate.ndim == 2 and estimate.size(1) > 1:
                estimate = estimate[:, sorted_unique_indices]

        return estimate, new_time, weight, weight_new_time

    @staticmethod
    def _update_auc_estimate(
        estimate: torch.Tensor, new_time: torch.Tensor
    ) -> torch.Tensor:
        # squeeze estimate if shape = (n_samples, 1)
        if estimate.ndim == 2 and estimate.shape[1] == 1:
            estimate = estimate.squeeze(1)

        # Ensure estimate is (n_samples, n_times) shape
        if estimate.ndim == 1:
            estimate = estimate.unsqueeze(1).expand(
                (estimate.shape[0], new_time.shape[0])
            )

        return estimate

    @staticmethod
    def _update_auc_weight(
        time: torch.Tensor,
        new_time: torch.Tensor,
        weight: torch.Tensor,
        weight_new_time: torch.Tensor,
    ) -> torch.Tensor:
        # if weight was not specified, weight of 1
        if weight is None:
            weight = torch.ones_like(time)
            weight_new_time = torch.ones_like(new_time)

        return weight, weight_new_time


if __name__ == "__main__":
    import doctest

    # Run doctest
    results = doctest.testmod()
    if results.failed == 0:
        print("All tests passed.")
    else:
        print("Some doctests failed.")
        sys.exit(1)



import os
import unittest

import torch
from torch import nn

from torchsurv.loss.cox import neg_partial_log_likelihood
from torchsurv.loss.momentum import Momentum
from torchsurv.loss.weibull import neg_log_likelihood

# set seed for reproducibility
torch.manual_seed(42)


class TestMometum(unittest.TestCase):
    def test_momentum_weibull(self):
        model = Momentum(
            backbone=nn.Sequential(nn.Linear(8, 2)),  # Weibull expect two outputs
            loss=neg_log_likelihood,
        )
        x = torch.randn((3, 8))
        results = model.infer(x)
        self.assertEqual(results.size(), (3, 2))
        self.assertEqual(results.requires_grad, False)
        self.assertEqual(model.online.training, True)
        self.assertEqual(model.target.training, False)
        self.assertTrue(torch.equal(results, model.target(x)))

    def test_momentum_cox(self):
        model = Momentum(
            backbone=nn.Sequential(nn.Linear(8, 1)),  # Cox expect one outputs
            loss=neg_partial_log_likelihood,
        )
        x = torch.randn((3, 8))
        results = model.infer(x)
        self.assertEqual(results.size(), (3, 1))
        self.assertEqual(results.requires_grad, False)
        self.assertEqual(model.online.training, True)
        self.assertEqual(model.target.training, False)
        self.assertTrue(torch.equal(results, model.target(x)))


if __name__ == "__main__":
    unittest.main()



import torch


def validate_log_shape(log_params: torch.Tensor) -> torch.Tensor:
    """Private function, check if the log shape is missing and impute it with 0
    if needed.
    Used only for Weibull loss, as it can handle either log_scale alone or both
    log_scale and log_shape."""
    if any(
        [
            log_params.dim() == 0,
            log_params.dim() == 1,  # if shape = [n_samples]
            log_params.dim() > 1
            and log_params.size(1) == 1,  # if shape = [n_samples, 1]
        ]
    ):
        if log_params.dim() == 1:
            log_params = log_params.unsqueeze(1)

        # Missing log shape parameter. Creating zeros placeholder instead.
        log_params = torch.hstack((log_params, torch.zeros_like(log_params)))

    return log_params


def check_within_follow_up(
    new_time: torch.Tensor, time: torch.Tensor, within_follow_up: bool
) -> None:
    # Check if the within_follow_up flag is set to True
    if within_follow_up:
        # Check if any value in new_time is outside the range of time
        if new_time.max() >= time.max() or new_time.min() < time.min():
            # Get the minimum and maximum values of time
            min_time = time.min().item()
            max_time = time.max().item()
            # Raise a ValueError if new_time is not within the follow-up time range
            raise ValueError(
                f"Value error: All new_time must be within follow-up time of test data: [{min_time}; {max_time}"
            )


def validate_new_time(
    new_time: torch.Tensor, time: torch.Tensor, within_follow_up: bool = True
) -> None:
    """
    Validate the new_time tensor for survival analysis functions.

    Args:
        new_time (torch.Tensor, float): Time points for metric computation of size n_times.
        time (torch.Tensor, float): Event or censoring time of size n_samples.
        within_follow_up (bool, optional): Whether values of ``new_time`` must be within values in ``time``. Defaults to True.

    Raises:
        ValueError: If ``new_time`` contains duplicate values.
        ValueError: If ``new_time`` is not sorted.
        TypeError: If ``new_time`` is not a tensor.
        ValueError: If ``new_time`` is not of floating-point type.
        ValueError: If ``new_time`` is not within the range of follow-up in ``time``. Assessed only if ``within_follow_up`` is True.
    """
    if not isinstance(new_time, torch.Tensor):
        raise TypeError("Type error: Input 'new_time' should be a tensor.")

    if not torch.is_floating_point(new_time):
        raise ValueError(
            "Value error: Input 'new_time' should be of floating-point type."
        )

    new_time_sorted, _ = torch.sort(new_time)
    if not torch.equal(new_time_sorted, new_time):
        raise ValueError(
            "Value error: Input 'new_time' should be sorted from the smallest time to the largest."
        )

    if len(new_time_sorted) != len(torch.unique(new_time_sorted)):
        raise ValueError("Value error: Input 'new_time' should contain unique values.")

    check_within_follow_up(new_time, time, within_follow_up)


def validate_survival_data(event: torch.Tensor, time: torch.Tensor):
    """Perform format and validity checks for survival data.

    Args:
        event (torch.Tensor, boolean):
            Event indicator of size n_samples (= True if event occurred).
        time (torch.Tensor, float):
            Event or censoring time of size n_samples.

    Raises:
        TypeError: If ``event`` or ``time`` are not tensors.
        ValueError: If ``event`` is not boolean.
        ValueError: If ``event`` and ``time`` are not of the same length.
        ValueError: If all ``event`` are False.
        ValueError: If any ``time`` are negative.
    """
    validate_tensor(event, "event")
    validate_tensor(time, "time")

    if not event.dtype == torch.bool:
        raise ValueError("Input 'event' should be of boolean type.")

    if not torch.is_floating_point(time):
        raise ValueError("Input 'time' should be of float type.")

    if len(event) != len(time):
        raise ValueError(
            "Dimension mismatch: Incompatible length between inputs 'time' and 'event'."
        )

    if torch.sum(event) <= 0:
        raise ValueError("All samples are censored.")

    if torch.any(time < 0.0):
        raise ValueError("Input 'time' should be non-negative.")


def validate_tensor(tensor: torch.Tensor, name: str) -> None:
    if not isinstance(tensor, torch.Tensor):
        raise TypeError(f"Input '{name}' should be a tensor")


def validate_event(event: torch.Tensor) -> torch.Tensor:
    if event.dtype == torch.bool:
        event = event.float()  # convert boolean to float for internal computation
    if not torch.all((event == 0) | (event == 1)):
        raise ValueError("Invalid values: 'event' must contain only [0, 1] values")
    return event


def validate_model_type(log_params: torch.Tensor, model_type: str) -> None:
    if model_type.lower() == "weibull":
        if log_params.dim() not in [1, 2]:
            raise ValueError(
                f"For Weibull model, 'log_params' must have shape (n_samples, 2) or (n_samples, 1). Found {log_params.dim()} dimensions."
            )
    elif model_type.lower() == "cox":
        if log_params.dim() != 1:
            raise ValueError(
                "For Cox model, 'log_params' must have shape (n_samples, 1)."
            )
    else:
        raise ValueError("Invalid model type. Must be 'weibull' or 'cox'.")


def validate_loss(
    log_params: torch.Tensor, event: torch.Tensor, time: torch.Tensor, model_type: str
) -> None:
    # sanity checks
    validate_tensor(log_params, "log_params")
    validate_tensor(event, "event")
    validate_tensor(time, "time")

    log_params = log_params.squeeze()
    event = event.squeeze()
    time = time.squeeze()

    if log_params.shape[0] != len(event):
        raise ValueError(
            "Dimension mismatch: 'log_params' and 'event' must have the same length"
        )

    if time.shape != event.shape:
        raise ValueError(
            "Dimension mismatch: 'time' and 'event' must have the same shape"
        )

    if torch.any(time < 0):
        raise ValueError("All elements in 'time' must be non-negative.")

    event = validate_event(event)
    validate_model_type(log_params, model_type)


if __name__ == "__main__":
    log_params_weibull = torch.randn((5, 2))
    log_params_cox = torch.randn((5, 1))
    event_data = torch.tensor([1, 0, 1, 1, 0])
    time_data = torch.tensor([10, 20, 30, 40, 50])

    # Validate Weibull model inputs
    validate_loss(log_params_weibull, event_data, time_data, model_type="weibull")

    # Validate Cox model inputs
    validate_loss(log_params_cox, event_data, time_data, model_type="cox")

    # Valid booleans values
    validate_loss(log_params_cox, event_data.bool(), time_data, model_type="cox")

    # check that the output is a ValueError
    try:
        validate_loss(
            log_params_weibull,
            torch.tensor([1, 0, 1]),
            time_data,
            model_type="weibull",
        )
    except ValueError as e:
        print(f"ValueError: {e}")



# global modules
import json
import unittest

import numpy as np
import torch
from sksurv.metrics import cumulative_dynamic_auc
from utils import DataBatchContainer, conditions_ci, conditions_p_value

# Local modules
from torchsurv.metrics.auc import Auc
from torchsurv.stats.ipcw import get_ipcw

# set seed for reproducibility
torch.manual_seed(42)
np.random.seed(42)

# Load the benchmark metrics from R
with open("tests/benchmark_data/benchmark_auc.json") as file:
    benchmark_aucs = json.load(file)

auc = Auc()


class TestAUC(unittest.TestCase):
    """_summary_
    List of packages compared
        - survAUC (R): AUC C/D with IPCW
                       Integral of AUC C/D
                       Integral of AUC I/D
        - RiskRegression (R): AUC C/D with IPCW
                              Standard error of AUC C/D with IPCW
        - timeROC (R): AUC C/D with IPCW
                       Standard error of AUC C/D with IPCW
        - sksurv (Python): AUC C/D with IPCW
    """

    def test_auc_cd_real_data(self):
        """test point estimate of auc cumulative/dynamic with ipcw on lung and gbsg datasets"""
        for benchmark_auc in benchmark_aucs:
            train_event = torch.tensor(benchmark_auc["train_status"]).bool()
            train_time = torch.tensor(benchmark_auc["train_time"], dtype=torch.float32)
            test_event = torch.tensor(benchmark_auc["test_status"]).bool()
            test_time = torch.tensor(benchmark_auc["test_time"], dtype=torch.float32)
            estimate = torch.tensor(benchmark_auc["estimate"], dtype=torch.float32)
            new_time = torch.tensor(benchmark_auc["times"], dtype=torch.float32)

            # ipcw obtained using censoring distribution estimated on train data
            ipcw = get_ipcw(train_event, train_time, test_time)
            ipcw_new_time = get_ipcw(train_event, train_time, new_time)
            auc_cd = auc(
                estimate,
                test_event,
                test_time,
                auc_type="cumulative",
                weight=ipcw,
                weight_new_time=ipcw_new_time,
                new_time=new_time,
            )

            auc_cd_survAUC = benchmark_auc["auc_cd_survAUC"]  # survAUC
            auc_cd_Uno_riskRegression = benchmark_auc[
                "auc_cd_Uno_riskRegression"
            ]  # riskRegression

            self.assertTrue(
                np.allclose(
                    auc_cd.numpy(), np.array(auc_cd_survAUC), rtol=1e-1, atol=1e-8
                )
            )
            self.assertTrue(
                np.allclose(
                    auc_cd.numpy(),
                    np.array(auc_cd_Uno_riskRegression),
                    rtol=1e-1,
                    atol=1e-2,
                )
            )

            # commented out: riskRegression does not match other R packages
            # auc_cd_se = auc._auc_se()

            # auc_cd_Uno_se_riskRegression = benchmark_auc[
            #     "auc_cd_Uno_se_riskRegression"
            # ]  # riskRegression

            # self.assertTrue(
            #     np.allclose(
            #         auc_cd_se.numpy(),
            #         np.array(auc_cd_Uno_se_riskRegression),
            #         rtol=1e-1,
            #         atol=1e-1,
            #     )
            # )

    def test_auc_cd_se_real_data(self):
        """test standard error of auc cumulative/dynamic with ipcw on lung and gbsg datasets"""
        for benchmark_auc in benchmark_aucs:
            test_event = torch.tensor(benchmark_auc["test_status"]).bool()
            test_time = torch.tensor(benchmark_auc["test_time"], dtype=torch.float32)
            estimate = torch.tensor(benchmark_auc["estimate"], dtype=torch.float32)
            new_time = torch.tensor(benchmark_auc["times_se"], dtype=torch.float32)

            # ipcw obtained using censoring distribution estimated on test data
            ipcw = get_ipcw(test_event, test_time, test_time)
            ipcw_new_time = get_ipcw(test_event, test_time, new_time)
            auc_cd = auc(
                estimate,
                test_event,
                test_time,
                auc_type="cumulative",
                weight=ipcw,
                weight_new_time=ipcw_new_time,
                new_time=new_time,
            )  # point estimate
            auc_cd_se = auc._auc_se()  # standard error

            # timeROC
            auc_cd_Uno_timeROC = benchmark_auc["auc_cd_Uno_timeROC"]  # point estimate
            auc_cd_Uno_se_timeROC = benchmark_auc[
                "auc_cd_Uno_se_timeROC"
            ]  # standard error

            self.assertTrue(
                np.allclose(
                    auc_cd.numpy(),
                    np.array(auc_cd_Uno_timeROC),
                    rtol=1e-1,
                    atol=1e-8,
                )
            )

            self.assertTrue(
                np.allclose(
                    auc_cd_se.numpy(),
                    np.array(auc_cd_Uno_se_timeROC),
                    rtol=1e-1,
                    atol=1e-8,
                )
            )

    def test_i_auc_real_data(self):
        """test integral of auc with survAUC on lung and gbsg datasets"""
        for benchmark_auc in benchmark_aucs:
            times = torch.tensor(benchmark_auc["times"], dtype=torch.float64)
            S = torch.tensor(benchmark_auc["surv.prob"], dtype=torch.float64)

            # integral of auc cumulative/dynamic
            auc_cd_survAUC = torch.tensor(
                benchmark_auc["auc_cd_survAUC"], dtype=torch.float64
            )
            auc_ne = Auc()
            auc_ne.auc = auc_cd_survAUC
            auc_ne.new_time = times
            i_auc_cd = auc_ne._integrate_cumulative(S, times.max())

            i_auc_cd_survAUC = benchmark_auc["iauc_cd_survAUC"]  # survAUC

            self.assertTrue(
                np.isclose(
                    i_auc_cd.numpy(), np.array(i_auc_cd_survAUC), rtol=1e-3, atol=1e-8
                )
            )

            # integral of auc incident/dynamic
            auc_id_sz_survAUC = torch.tensor(
                benchmark_auc["auc_id_sz_survAUC"], dtype=torch.float64
            )
            auc_ne = Auc()
            auc_ne.auc = auc_id_sz_survAUC
            auc_ne.new_time = times
            i_auc_id_sz = auc_ne._integrate_incident(S, times.max())
            i_auc_id_sz_survAUC = benchmark_auc["i_auc_id_sz_survAUC"]  # survAUC

            self.assertTrue(
                np.isclose(
                    i_auc_id_sz.numpy(),
                    np.array(i_auc_id_sz_survAUC),
                    rtol=1e-3,
                    atol=1e-8,
                )
            )

    def test_auc_cd_simulated_data(self):
        """test auc cumulative/dynamic on simulated batches including edge cases"""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=20,
            flags_to_set=[
                "train_ties_time_event",
                "test_ties_time_event",
                "train_ties_time_censoring",
                "test_ties_time_censoring",
                "train_ties_time_event_censoring",
                "test_ties_time_event_censoring",
                "test_no_censoring",
                #'train_no_censoring', cumulative_dynamic_auc from sksurv fails
                "test_event_at_last_time",
                "ties_score_events",
                "ties_score_event_censoring",
                "ties_score_censoring",
            ],
        )
        for batch in batch_container.batches:
            (
                train_time,
                train_event,
                test_time,
                test_event,
                estimate,
                new_time,
                y_train_array,
                y_test_array,
                _,
                new_time_array,
            ) = batch

            ipcw = get_ipcw(train_event, train_time, test_time)
            ipcw_new_time = get_ipcw(train_event, train_time, new_time)

            # expand in 2D
            estimate = estimate.unsqueeze(1).expand((len(test_time), len(new_time)))
            estimate = estimate + torch.randn_like(estimate) * 0.1

            auc_cd = auc(
                estimate,
                test_event,
                test_time,
                auc_type="cumulative",
                weight=ipcw,
                weight_new_time=ipcw_new_time,
                new_time=new_time,
            )

            auc_cd_sksurv, _ = cumulative_dynamic_auc(
                y_train_array, y_test_array, estimate.numpy(), new_time_array
            )  # sksurv

            self.assertTrue(
                np.allclose(auc_cd.numpy(), auc_cd_sksurv, rtol=1e-5, atol=1e-8)
            )

    def test_auc_confidence_interval_pvalue(self):
        """test auc confidence interval and p-value are as expected"""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=20,
            flags_to_set=[
                "test_ties_time_event",
                "test_ties_time_censoring",
                "test_ties_time_event_censoring",
                #'test_no_censoring', concordance_index_censored fails
                "test_event_at_last_time",
                "ties_score_events",
                "ties_score_event_censoring",
                "ties_score_censoring",
            ],
        )
        for batch in batch_container.batches:
            (_, _, time, event, estimate, *_) = batch

            # Run c-index
            auc(
                estimate,
                event,
                time,
            )

            n_bootstraps = 9

            for method in ["blanche", "bootstrap"]:
                for alternative in ["two_sided", "less", "greater"]:
                    auc_ci = auc.confidence_interval(
                        method=method,
                        alternative=alternative,
                        n_bootstraps=n_bootstraps,
                    )
                    self.assertTrue(
                        all([conditions_ci(auc_ci[:, i]) for i in range(len(auc.auc))])
                    )

            for method in ["blanche", "bootstrap"]:
                for alternative in ["two_sided", "less", "greater"]:
                    auc_p_value = auc.p_value(
                        method=method,
                        alternative=alternative,
                        n_bootstraps=n_bootstraps,
                    )

                    self.assertTrue(
                        all(
                            [
                                conditions_p_value(auc_p_value[i])
                                for i in range(len(auc.auc))
                            ]
                        )
                    )

    def test_auc_compare(self):
        "test compare function of auc behavesas expected."
        _ = torch.manual_seed(42)
        n = 128
        estimate_informative = torch.randn(
            (n,)
        )  # estimate used to define time-to-event
        estimate_non_informative = torch.randn((n,))  # random estimate
        event = torch.randint(low=0, high=2, size=(n,)).bool()
        time = (
            torch.randn(size=(n,)) * 10 - estimate_informative * 5.0 + 200
        )  # + estimate for auc < 0.5 and - for auc > 0.5

        Auc_informative = Auc()
        auc_informative = Auc_informative(estimate_informative, event, time)

        Auc_non_informative = Auc()
        auc_non_informative = Auc_non_informative(estimate_non_informative, event, time)

        p_value_compare_informative = Auc_informative.compare(Auc_non_informative)
        p_value_compare_non_informative = Auc_non_informative.compare(Auc_informative)

        self.assertTrue(np.all(auc_informative.numpy() > auc_non_informative.numpy()))
        self.assertTrue(np.any(p_value_compare_informative.numpy() < 0.05))
        self.assertTrue(np.all(p_value_compare_non_informative.numpy() > 0.05))

    def test_auc_error_raised(self):
        """test that errors are raised in not-accepted edge cases."""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=2,
            flags_to_set=[
                "test_all_censored",
                "test_max_time_in_new_time",
            ],
        )
        for batch in batch_container.batches:
            _, _, test_time, test_event, estimate, new_time, *_ = batch

            self.assertRaises(
                ValueError,
                auc,
                estimate,
                test_event,
                test_time,
                new_time=new_time,
            )


if __name__ == "__main__":
    unittest.main()



import json
import unittest

import numpy as np
import torch
from sksurv.metrics import CensoringDistributionEstimator, SurvivalFunctionEstimator
from utils import DataBatchContainer

# local
from torchsurv.stats.kaplan_meier import KaplanMeierEstimator

# Load the benchmark cox log likelihoods from R
with open("tests/benchmark_data/benchmark_kaplan_meier.json") as file:
    benchmark_kaplan_meiers = json.load(file)

# set seed for reproducibility
torch.manual_seed(42)
np.random.seed(42)


class TestNonParametric(unittest.TestCase):
    """
    List of packages compared
        - survival (R)
        - sksurv (Python)
    """

    def test_kaplan_meier_survival_distribution_real_data(self):
        """test Kaplan Meier survival distribution estimate on lung and gbsg datasets"""

        for benchmark_kaplan_meier in benchmark_kaplan_meiers:
            event = torch.tensor(benchmark_kaplan_meier["status"]).bool()
            time = torch.tensor(benchmark_kaplan_meier["time"], dtype=torch.float32)
            new_time = torch.tensor(
                benchmark_kaplan_meier["times"], dtype=torch.float32
            )

            km = KaplanMeierEstimator()
            km(event, time, censoring_dist=False)
            st = km.predict(new_time)

            st_survival = np.array(benchmark_kaplan_meier["surv_prob_survival"])

            self.assertTrue(
                np.allclose(
                    st.numpy(),
                    st_survival,
                    rtol=1e-3,
                    atol=1e-8,
                )
            )

    def test_kaplan_meier_censoring_distribution_simulated_data(self):
        """test Kaplan Meier estimate of censoring distribution on simulated batches including edge cases"""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=20,
            flags_to_set=[
                "train_ties_time_event",
                "train_ties_time_censoring",
                "train_ties_time_event_censoring",
                "train_no_censoring",
            ],
        )
        for batch in batch_container.batches:
            (time, event, _, _, _, _, y_array, *_) = batch

            km = KaplanMeierEstimator()
            km(event, time, censoring_dist=True)
            ct = km.km_est

            cens = CensoringDistributionEstimator()
            cens.fit(y_array)
            ct_sksurv = cens.prob_
            if not event.all():
                ct_sksurv = ct_sksurv[1:]

            self.assertTrue(np.allclose(ct.numpy(), ct_sksurv, rtol=1e-5, atol=1e-8))

    def test_kaplan_meier_survival_distribution_simulated_data(self):
        """test Kaplan Meier estimate of survival distribution on simulated batches including edge cases"""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=20,
            flags_to_set=[
                "train_ties_time_event",
                "train_ties_time_censoring",
                "train_ties_time_event_censoring",
                "train_no_censoring",
            ],
        )
        for batch in batch_container.batches:
            (time, event, _, _, _, _, y_array, *_) = batch

            km = KaplanMeierEstimator()
            km(event, time, censoring_dist=False)
            st = km.km_est

            surv = SurvivalFunctionEstimator()
            surv.fit(y_array)
            st_sksurv = surv.prob_[1:]

            self.assertTrue(np.allclose(st.numpy(), st_sksurv, rtol=1e-5, atol=1e-8))

    def test_kaplan_meier_predict_censoring_distribution_simulated_data(self):
        """test Kaplan Meier prediction of censoring distribution on simulated batches including edge cases"""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=20,
            flags_to_set=[
                "train_ties_time_event",
                "test_ties_time_event",
                "train_ties_time_censoring",
                "test_ties_time_censoring",
                "train_ties_time_event_censoring",
                "test_ties_time_event_censoring",
                "test_no_censoring",
                #'train_no_censoring', CensoringDistributionEstimator from sksurv fails
                "test_event_at_last_time",
            ],
        )
        for batch in batch_container.batches:
            (
                train_time,
                train_event,
                test_time,
                _,
                _,
                _,
                y_train_array,
                y_test_array,
                *_,
            ) = batch

            km = KaplanMeierEstimator()
            km(train_event, train_time, censoring_dist=True)
            ct_pred = km.predict(test_time)

            cens = CensoringDistributionEstimator()
            cens.fit(y_train_array)
            ct_pred_sksurv = cens.predict_proba(y_test_array["futime"])

            self.assertTrue(
                np.allclose(ct_pred.numpy(), ct_pred_sksurv, rtol=1e-5, atol=1e-8)
            )

    def test_kaplan_meier_predict_survival_distribution_simulated_data(self):
        """test Kaplan Meier prediction of survival distribution on simulated batches including edge cases"""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=20,
            flags_to_set=[
                "train_ties_time_event",
                "test_ties_time_event",
                "train_ties_time_censoring",
                "test_ties_time_censoring",
                "train_ties_time_event_censoring",
                "test_ties_time_event_censoring",
                "test_no_censoring",
                #'train_no_censoring', CensoringDistributionEstimator from sksurv fails
                "test_event_at_last_time",
            ],
        )
        for batch in batch_container.batches:
            (
                train_time,
                train_event,
                test_time,
                _,
                _,
                _,
                y_train_array,
                y_test_array,
                *_,
            ) = batch

            km = KaplanMeierEstimator()
            km(train_event, train_time, censoring_dist=False)
            st_pred = km.predict(test_time)

            surv = SurvivalFunctionEstimator()
            surv.fit(y_train_array)
            st_pred_sksurv = surv.predict_proba(y_test_array["futime"])

            self.assertTrue(
                np.allclose(st_pred.numpy(), st_pred_sksurv, rtol=1e-5, atol=1e-8)
            )

    def test_kaplan_meier_estimate_error_raised(self):
        """test that errors are raised for estimation in not-accepted edge cases."""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=1,
            flags_to_set=["train_all_censored"],
        )
        for batch in batch_container.batches:
            (train_time, train_event, *_) = batch

            self.assertRaises(
                ValueError, KaplanMeierEstimator(), train_event, train_time
            )

    def test_kaplan_meier_prediction_error_raised(self):
        """test that errors are raised for prediction in not-accepted edge cases."""
        batch_container = DataBatchContainer()
        batch_container.generate_batches(
            n_batch=1,
            flags_to_set=["test_max_time_gt_train_max_time"],
        )
        for batch in batch_container.batches:
            (train_time, train_event, test_time, *_) = batch

            train_event[
                -1
            ] = False  # if last event is censoring, the last KM is > 0 and it cannot predict beyond this time
            km = KaplanMeierEstimator()
            km(train_event, train_time, censoring_dist=False)

            self.assertRaises(ValueError, km.predict, test_time)

            def test_kaplan_meier_plot_km(self):
                """test Kaplan Meier plot function"""
                import matplotlib.pyplot as plt

                event = torch.tensor([1, 1, 0, 1, 0, 1, 0, 0, 1, 1], dtype=torch.bool)
                time = torch.tensor([1, 2, 2, 3, 3, 4, 4, 5, 6, 7], dtype=torch.float32)

                km = KaplanMeierEstimator()
                km(event, time, censoring_dist=False)

                fig, ax = plt.subplots()
                km.plot_km(ax=ax)
                plt.close(fig)  # Close the plot to avoid displaying it during tests

            def test_kaplan_meier_print_survival_table(self):
                """test Kaplan Meier print survival table function"""
                event = torch.tensor([1, 1, 0, 1, 0, 1, 0, 0, 1, 1], dtype=torch.bool)
                time = torch.tensor([1, 2, 2, 3, 3, 4, 4, 5, 6, 7], dtype=torch.float32)

                km = KaplanMeierEstimator()
                km(event, time, censoring_dist=False)
                km.print_survival_table()

                # Check if the survival table is printed correctly
                expected_output = (
                    "Time\tSurvival\n"
                    "----------------\n"
                    "1.00\t1.0000\n"
                    "2.00\t0.8889\n"
                    "3.00\t0.6667\n"
                    "4.00\t0.5000\n"
                    "5.00\t0.5000\n"
                    "6.00\t0.3333\n"
                    "7.00\t0.0000\n"
                )
                with self.assertLogs(level="INFO") as log:
                    km.print_survival_table()
                    self.assertIn(expected_output, log.output)


if __name__ == "__main__":
    unittest.main()
